@@ -1,14 +1,13 @@
 #include <stdlib.h>
-#include <string.h>
 #include <assert.h>
 #include <link.h>
+#include <gelf.h>
 
 #ifndef EF_ARM_ABI_FLOAT_HARD
 # define EF_ARM_ABI_FLOAT_HARD  EF_ARM_VFP_FLOAT
 #endif
 
-#include "mcount-arch.h"
-#include "libmcount/mcount.h"
+#include "libmcount/internal.h"
 #include "utils/utils.h"
 #include "utils/symbol.h"
 #include "utils/rbtree.h"
@@ -22,8 +21,17 @@ struct lr_offset {
 #define REG_SP  13
 
 static struct rb_root offset_cache = RB_ROOT;
+
+/* whether current machine supports hardfp */
 static bool use_hard_float = false;
+
+#ifdef HAVE_ARM_HARDFP
+/* need to check hardfp at runtime */
 static bool float_abi_checked = false;
+#else
+/* disable hardfp as it's not supported */
+static bool float_abi_checked = true;
+#endif
 
 struct offset_entry {
 	struct rb_node node;
@@ -305,7 +313,7 @@ void check_float_abi(void)
 }
 
 int mcount_get_register_arg(struct mcount_arg_context *ctx,
-			    struct ftrace_arg_spec *spec)
+			    struct uftrace_arg_spec *spec)
 {
 	struct mcount_regs *regs = ctx->regs;
 	int reg_idx;
@@ -350,6 +358,8 @@ int mcount_get_register_arg(struct mcount_arg_context *ctx,
 	case ARM_REG_R3:
 		ctx->val.i = ARG4(regs);
 		break;
+
+#ifdef HAVE_ARM_HARDFP
 	case ARM_REG_S0:
 		asm volatile ("vstr %%s0, %0\n" : "=m" (ctx->val.v));
 		break;
@@ -422,6 +432,8 @@ int mcount_get_register_arg(struct mcount_arg_context *ctx,
 	case ARM_REG_D7:
 		asm volatile ("vstr %%d7, %0\n" : "=m" (ctx->val.v));
 		break;
+#endif /* HAVE_ARM_HARDFP */
+
 	default:
 		return -1;
 	}
@@ -430,7 +442,7 @@ int mcount_get_register_arg(struct mcount_arg_context *ctx,
 }
 
 void mcount_get_stack_arg(struct mcount_arg_context *ctx,
-			  struct ftrace_arg_spec *spec)
+			  struct uftrace_arg_spec *spec)
 {
 	int offset = 1;
 
@@ -466,7 +478,7 @@ void mcount_get_stack_arg(struct mcount_arg_context *ctx,
 }
 
 void mcount_arch_get_arg(struct mcount_arg_context *ctx,
-			 struct ftrace_arg_spec *spec)
+			 struct uftrace_arg_spec *spec)
 {
 	if (!float_abi_checked)
 		check_float_abi();
@@ -480,7 +492,7 @@ void mcount_arch_get_arg(struct mcount_arg_context *ctx,
 }
 
 void mcount_arch_get_retval(struct mcount_arg_context *ctx,
-			    struct ftrace_arg_spec *spec)
+			    struct uftrace_arg_spec *spec)
 {
 	if (!float_abi_checked)
 		check_float_abi();
@@ -490,6 +502,7 @@ void mcount_arch_get_retval(struct mcount_arg_context *ctx,
 		spec->size = 8;
 
 	/* type of return value cannot be FLOAT, so check format instead */
+#ifdef HAVE_ARM_HARDFP
 	if (spec->fmt == ARG_FMT_FLOAT && use_hard_float) {
 		if (spec->size <= 4)
 			asm volatile ("vstr %%s0, %0\n" : "=m" (ctx->val.v));
@@ -497,5 +510,11 @@ void mcount_arch_get_retval(struct mcount_arg_context *ctx,
 			asm volatile ("vstr %%d0, %0\n" : "=m" (ctx->val.v));
 	}
 	else
+#endif /* HAVE_ARM_HARDFP */
 		memcpy(ctx->val.v, ctx->retval, spec->size);
 }
+
+unsigned long mcount_arch_plthook_addr(struct plthook_data *pd, int idx)
+{
+	return pd->plt_addr;
+}
