@@ -1,5 +1,4 @@
 #include <stdio.h>
-#include <string.h>
 #include <stdint.h>
 #include <stdbool.h>
 #include <inttypes.h>
@@ -8,12 +7,13 @@
 #include <sys/stat.h>
 
 #include "uftrace.h"
-#include "utils/compiler.h"
 #include "utils/list.h"
 #include "utils/utils.h"
 #include "utils/fstack.h"
 #include "utils/filter.h"
+#include "utils/kernel.h"
 #include "libtraceevent/kbuffer.h"
+#include "libtraceevent/event-parse.h"
 
 
 struct uftrace_dump_ops {
@@ -31,17 +31,28 @@ struct uftrace_dump_ops {
 			    struct ftrace_task_handle *task, char *name);
 	/* this is called when kernel data starts */
 	void (*kernel_start)(struct uftrace_dump_ops *ops,
-			     struct ftrace_kernel *kernel);
+			     struct uftrace_kernel_reader *kernel);
 	/* this is called when a cpu data start */
 	void (*cpu_start)(struct uftrace_dump_ops *ops,
-			  struct ftrace_kernel *kernel, int cpu);
+			  struct uftrace_kernel_reader *kernel, int cpu);
 	/* this is called for each kernel-level function entry/exit */
-	void (*kernel)(struct uftrace_dump_ops *ops,
-		       struct ftrace_kernel *kernel, int cpu,
-		       struct ftrace_ret_stack *frs, char *name);
+	void (*kernel_func)(struct uftrace_dump_ops *ops,
+			    struct uftrace_kernel_reader *kernel, int cpu,
+			    struct uftrace_record *frs, char *name);
+	/* this is called for each kernel event (tracepoint) */
+	void (*kernel_event)(struct uftrace_dump_ops *ops,
+			     struct uftrace_kernel_reader *kernel, int cpu,
+			     struct uftrace_record *frs);
 	/* thius is called when there's a lost record (usually in kernel) */
 	void (*lost)(struct uftrace_dump_ops *ops,
 		     uint64_t time, int tid, int losts);
+	/* this is called when a perf data (for each cpu) starts */
+	void (*perf_start)(struct uftrace_dump_ops *ops,
+			   struct uftrace_perf_reader *perf, int cpu);
+	/* this is called for each perf event (e.g. schedule) */
+	void (*perf_event)(struct uftrace_dump_ops *ops,
+			   struct uftrace_perf_reader *perf,
+			   struct uftrace_record *frs);
 	/* this is called at the end */
 	void (*footer)(struct uftrace_dump_ops *ops,
 		       struct ftrace_file_handle *handle, struct opts *opts);
@@ -66,10 +77,17 @@ struct uftrace_flame_dump {
 	uint64_t sample_time;
 };
 
+static const char * rstack_type(struct uftrace_record *frs)
+{
+	return frs->type == UFTRACE_EXIT ? "exit " :
+		frs->type == UFTRACE_ENTRY ? "entry" :
+		frs->type == UFTRACE_EVENT ? "event" : "lost ";
+}
+
 static void pr_time(uint64_t timestamp)
 {
-	unsigned sec   = timestamp / 1000000000;
-	unsigned nsec  = timestamp % 1000000000;
+	unsigned sec   = timestamp / NSEC_PER_SEC;
+	unsigned nsec  = timestamp % NSEC_PER_SEC;
 
 	pr_out("%u.%09u  ", sec, nsec);
 }
@@ -78,10 +96,10 @@ static int pr_task(struct opts *opts)
 {
 	FILE *fp;
 	char buf[PATH_MAX];
-	struct ftrace_msg msg;
-	struct ftrace_msg_task tmsg;
-	struct ftrace_msg_sess smsg;
-	char *exename;
+	struct uftrace_msg msg;
+	struct uftrace_msg_task tmsg;
+	struct uftrace_msg_sess smsg;
+	char *exename = NULL;
 
 	snprintf(buf, sizeof(buf), "%s/task", opts->dirname);
 	fp = fopen(buf, "r");
@@ -89,13 +107,13 @@ static int pr_task(struct opts *opts)
 		return -1;
 
 	while (fread(&msg, sizeof(msg), 1, fp) == 1) {
-		if (msg.magic != FTRACE_MSG_MAGIC) {
+		if (msg.magic != UFTRACE_MSG_MAGIC) {
 			pr_red("invalid message magic: %hx\n", msg.magic);
 			goto out;
 		}
 
 		switch (msg.type) {
-		case FTRACE_MSG_TID:
+		case UFTRACE_MSG_TASK_START:
 			if (fread(&tmsg, sizeof(tmsg), 1, fp) != 1) {
 				pr_red("cannot read task message: %m\n");
 				goto out;
@@ -104,7 +122,7 @@ static int pr_task(struct opts *opts)
 			pr_time(tmsg.time);
 			pr_out("task tid %d (pid %d)\n", tmsg.tid, tmsg.pid);
 			break;
-		case FTRACE_MSG_FORK_END:
+		case UFTRACE_MSG_FORK_END:
 			if (fread(&tmsg, sizeof(tmsg), 1, fp) != 1) {
 				pr_red("cannot read task message: %m\n");
 				goto out;
@@ -113,11 +131,13 @@ static int pr_task(struct opts *opts)
 			pr_time(tmsg.time);
 			pr_out("fork pid %d (ppid %d)\n", tmsg.tid, tmsg.pid);
 			break;
-		case FTRACE_MSG_SESSION:
+		case UFTRACE_MSG_SESSION:
 			if (fread(&smsg, sizeof(smsg), 1, fp) != 1) {
 				pr_red("cannot read session message: %m\n");
 				goto out;
 			}
+
+			free(exename);
 			exename = xmalloc(ALIGN(smsg.namelen, 8));
 			if (fread(exename, ALIGN(smsg.namelen, 8), 1,fp) != 1 ) {
 				pr_red("cannot read executable name: %m\n");
@@ -127,7 +147,6 @@ static int pr_task(struct opts *opts)
 			pr_time(smsg.task.time);
 			pr_out("session of task %d: %.*s (%s)\n",
 			       smsg.task.tid, sizeof(smsg.sid), smsg.sid, exename);
-			free(exename);
 			break;
 		default:
 			pr_out("unknown message type: %u\n", msg.type);
@@ -136,6 +155,7 @@ static int pr_task(struct opts *opts)
 	}
 
 out:
+	free(exename);
 	fclose(fp);
 	return 0;
 }
@@ -210,7 +230,8 @@ static int pr_task_txt(struct opts *opts)
 			}
 			*end++ = '\0';
 
-			sscanf(end, "tid=%d sid=%s", &tid, sid);
+			if (sscanf(end, "pid=%d sid=%s", &tid, sid) != 2)
+				sscanf(end, "tid=%d sid=%s", &tid, sid);
 
 			ptr = strstr(end, "exename=");
 			if (ptr == NULL) {
@@ -280,7 +301,7 @@ static void pr_hex(uint64_t *offset, void *data, size_t len)
 
 static void pr_args(struct fstack_arguments *args)
 {
-	struct ftrace_arg_spec *spec;
+	struct uftrace_arg_spec *spec;
 	void *ptr = args->data;
 	size_t size;
 	int i = 0;
@@ -290,25 +311,63 @@ static void pr_args(struct fstack_arguments *args)
 		if (spec->idx == RETVAL_IDX)
 			continue;
 
-		if (spec->fmt == ARG_FMT_STR) {
-			char buf[64];
+		if (spec->fmt == ARG_FMT_STR ||
+		    spec->fmt == ARG_FMT_STD_STRING) {
+			char *buf;
 			const int null_str = -1;
 
 			size = *(unsigned short *)ptr;
+			buf = xmalloc(size + 1);
 			strncpy(buf, ptr + 2, size);
+			buf[size] = '\0';
 
 			if (!memcmp(buf, &null_str, 4))
 				strcpy(buf, "NULL");
 
-			pr_out("  args[%d] str: %s\n", i , buf);
+			if (spec->fmt == ARG_FMT_STD_STRING)
+				pr_out("  args[%d] std::string: %s\n", i , buf);
+			else
+				pr_out("  args[%d] str: %s\n", i , buf);
+
+			free(buf);
 			size += 2;
 		}
+		else if (spec->fmt == ARG_FMT_FUNC_PTR) {
+			struct ftrace_task_handle *task;
+			struct uftrace_session_link *sessions;
+			struct sym *sym;
+			unsigned long val = 0;
+
+			task = container_of(args, struct ftrace_task_handle, args);
+			sessions = &task->h->sessions;
+
+			memcpy(&val, ptr, spec->size);
+			size = spec->size;
+
+			sym = task_find_sym_addr(sessions, task,
+						 task->rstack->time,
+						 (uint64_t)val);
+
+			if (sym)
+				pr_out("  args[%d] p: &%s\n", i, sym->name);
+			else
+				pr_out("  args[%d] p: %p\n", i, (void *)val);
+		}
+		else if (spec->fmt == ARG_FMT_ENUM) {
+			long long val = 0;
+
+			memcpy(&val, ptr, spec->size);
+			pr_out("  args[%d] enum %s: %s (%lld)\n", i,
+			       spec->enum_str, get_enum_string(spec->enum_str, val), val);
+			size = spec->size;
+		}
 		else {
 			long long val = 0;
 
 			memcpy(&val, ptr, spec->size);
-			pr_out("  args[%d] %c%d: %#llx\n", i,
-			       ARG_SPEC_CHARS[spec->fmt], spec->size * 8, val);
+			pr_out("  args[%d] %c%d: 0x%0*llx\n", i,
+			       ARG_SPEC_CHARS[spec->fmt], spec->size * 8,
+			       spec->size * 2, val);
 			size = spec->size;
 		}
 
@@ -319,18 +378,18 @@ static void pr_args(struct fstack_arguments *args)
 
 static void pr_retval(struct fstack_arguments *args)
 {
-	struct ftrace_arg_spec *spec;
+	struct uftrace_arg_spec *spec;
 	void *ptr = args->data;
 	size_t size;
-	int i = 0;
 
 	list_for_each_entry(spec, args->args, list) {
 		/* skip argument info */
 		if (spec->idx != RETVAL_IDX)
 			continue;
 
-		if (spec->fmt == ARG_FMT_STR) {
-			char buf[64];
+		if (spec->fmt == ARG_FMT_STR ||
+		    spec->fmt == ARG_FMT_STD_STRING) {
+			char buf[64] = {0};
 			const int null_str = -1;
 
 			size = *(unsigned short *)ptr;
@@ -339,20 +398,93 @@ static void pr_retval(struct fstack_arguments *args)
 			if (!memcmp(buf, &null_str, 4))
 				strcpy(buf, "NULL");
 
-			pr_out("  retval[%d] str: %s\n", i , buf);
+			if (spec->fmt == ARG_FMT_STD_STRING)
+				pr_out("  retval std::string: %s\n", buf);
+			else
+				pr_out("  retval str: %s\n", buf);
 			size += 2;
 		}
+		else if (spec->fmt == ARG_FMT_FUNC_PTR) {
+			struct ftrace_task_handle *task;
+			struct uftrace_session_link *sessions;
+			struct sym *sym;
+			unsigned long val = 0;
+
+			task = container_of(args, struct ftrace_task_handle, args);
+			sessions = &task->h->sessions;
+
+			memcpy(&val, ptr, spec->size);
+			size = spec->size;
+
+			sym = task_find_sym_addr(sessions, task,
+						 task->rstack->time,
+						 (uint64_t)val);
+
+			if (sym)
+				pr_out("  retval p: &%s\n", sym->name);
+			else
+				pr_out("  retval p: %p\n", (void *)val);
+		}
 		else {
 			long long val = 0;
 
 			memcpy(&val, ptr, spec->size);
-			pr_out("  retval[%d] %c%d: %#llx\n", i,
-			       ARG_SPEC_CHARS[spec->fmt], spec->size * 8, val);
+			pr_out("  retval %c%d: 0x%0*llx\n",
+			       ARG_SPEC_CHARS[spec->fmt], spec->size * 8,
+			       spec->size * 2, val);
 			size = spec->size;
 		}
 
 		ptr += ALIGN(size, 4);
-		i++;
+	}
+}
+
+static void pr_event(int eid, void *ptr, int len)
+{
+	union {
+		struct uftrace_proc_statm *statm;
+		struct uftrace_page_fault *pgfault;
+	} d;
+
+	/* built-in events */
+	switch (eid) {
+	case EVENT_ID_PROC_STATM:
+		d.statm = ptr;
+		pr_out("  proc/statm: vmsize=%"PRIu64"K vmrss=%"PRIu64"K shared=%"PRIu64"K\n",
+		       d.statm->vmsize, d.statm->vmrss, d.statm->shared);
+		break;
+	case EVENT_ID_PAGE_FAULT:
+		d.pgfault = ptr;
+		pr_out("  page-fault: major=%"PRIu64" minor=%"PRIu64"\n",
+		       d.pgfault->major, d.pgfault->minor);
+		break;
+	default:
+		break;
+	}
+
+	/* user events */
+}
+
+static void get_feature_string(char *buf, size_t sz, uint64_t feature_mask)
+{
+	int i;
+	size_t len;
+	bool first = true;
+	const char *feat_str[] = { "PLTHOOK", "TASK_SESSION", "KERNEL",
+				   "ARGUMENT", "RETVAL", "SYM_REL_ADDR",
+				   "MAX_STACK", "EVENT", "PERF_EVENT",
+				   "AUTO_ARGS" };
+
+	/* feat_str should match to enum uftrace_feat_bits */
+	for (i = 0; i < FEAT_BIT_MAX; i++) {
+		if (!((1U << i) & feature_mask))
+			continue;
+
+		len = snprintf(buf, sz, "%s%s", first ? "" : " | ", feat_str[i]);
+		buf += len;
+		sz  -= len;
+
+		first = false;
 	}
 }
 
@@ -361,6 +493,7 @@ static void print_raw_header(struct uftrace_dump_ops *ops,
 			     struct opts *opts)
 {
 	int i;
+	char buf[1024];
 	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
 
 	pr_out("uftrace file header: magic         = ");
@@ -373,7 +506,9 @@ static void print_raw_header(struct uftrace_dump_ops *ops,
 	       handle->hdr.endian, handle->hdr.endian == 1 ? "little" : "big");
 	pr_out("uftrace file header: class         = %u (%s bit)\n",
 	       handle->hdr.class, handle->hdr.class == 2 ? "64" : "32");
-	pr_out("uftrace file header: features      = %#"PRIx64"\n", handle->hdr.feat_mask);
+	get_feature_string(buf, sizeof(buf), handle->hdr.feat_mask);
+	pr_out("uftrace file header: features      = %#"PRIx64" (%s)\n",
+	       handle->hdr.feat_mask, buf);
 	pr_out("uftrace file header: info          = %#"PRIx64"\n", handle->hdr.info_mask);
 	pr_hex(&raw->file_offset, &handle->hdr, handle->hdr.header_size);
 	pr_out("\n");
@@ -392,7 +527,12 @@ static void print_raw_header(struct uftrace_dump_ops *ops,
 static void print_raw_task_start(struct uftrace_dump_ops *ops,
 				 struct ftrace_task_handle *task)
 {
+	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
+
 	pr_out("reading %d.dat\n", task->tid);
+	raw->file_offset = 0;
+
+	setup_rstack_list(&task->rstack_list);
 }
 
 static void print_raw_inverted_time(struct uftrace_dump_ops *ops,
@@ -408,42 +548,59 @@ static void print_raw_inverted_time(struct uftrace_dump_ops *ops,
 static void print_raw_task_rstack(struct uftrace_dump_ops *ops,
 				  struct ftrace_task_handle *task, char *name)
 {
-	struct ftrace_ret_stack *frs = task->rstack;
+	struct uftrace_record *frs = task->rstack;
 	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
 
+	if (frs->type == UFTRACE_EVENT)
+		name = get_event_name(task->h, frs->addr);
+
 	pr_time(frs->time);
 	pr_out("%5d: [%s] %s(%"PRIx64") depth: %u\n",
-	       task->tid, frs->type == FTRACE_EXIT ? "exit " :
-	       frs->type == FTRACE_ENTRY ? "entry" : "lost ",
+	       task->tid, rstack_type(frs),
 	       name, frs->addr, frs->depth);
 	pr_hex(&raw->file_offset, frs, sizeof(*frs));
 
+	if (frs->type == UFTRACE_EVENT)
+		free(name);
+
 	if (frs->more) {
-		if (frs->type == FTRACE_ENTRY) {
+		if (frs->type == UFTRACE_ENTRY) {
 			pr_time(frs->time);
 			pr_out("%5d: [%s] length = %d\n", task->tid, "args ",
 			       task->args.len);
 			pr_args(&task->args);
-			pr_hex(&raw->file_offset, task->args.data, task->args.len);
-		} else if (frs->type == FTRACE_EXIT) {
+			pr_hex(&raw->file_offset, task->args.data,
+			       ALIGN(task->args.len, 8));
+		}
+		else if (frs->type == UFTRACE_EXIT) {
 			pr_time(frs->time);
 			pr_out("%5d: [%s] length = %d\n", task->tid, "retval",
 			       task->args.len);
 			pr_retval(&task->args);
-			pr_hex(&raw->file_offset, task->args.data, task->args.len);
-		} else
+			pr_hex(&raw->file_offset, task->args.data,
+			       ALIGN(task->args.len, 8));
+		}
+		else if (frs->type == UFTRACE_EVENT) {
+			pr_time(frs->time);
+			pr_out("%5d: [%s] length = %d\n", task->tid, "data ",
+			       task->args.len);
+			pr_event(frs->addr, task->args.data, task->args.len);
+			pr_hex(&raw->file_offset, task->args.data,
+			       ALIGN(task->args.len, 8));
+		}
+		else
 			abort();
 	}
 }
 
 static void print_raw_kernel_start(struct uftrace_dump_ops *ops,
-				   struct ftrace_kernel *kernel)
+				   struct uftrace_kernel_reader *kernel)
 {
 	pr_out("\n");
 }
 
 static void print_raw_cpu_start(struct uftrace_dump_ops *ops,
-				struct ftrace_kernel *kernel, int cpu)
+				struct uftrace_kernel_reader *kernel, int cpu)
 {
 	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
 	struct kbuffer *kbuf = kernel->kbufs[cpu];
@@ -455,17 +612,46 @@ static void print_raw_cpu_start(struct uftrace_dump_ops *ops,
 }
 
 static void print_raw_kernel_rstack(struct uftrace_dump_ops *ops,
-				    struct ftrace_kernel *kernel, int cpu,
-				    struct ftrace_ret_stack *frs, char *name)
+				    struct uftrace_kernel_reader *kernel, int cpu,
+				    struct uftrace_record *frs, char *name)
 {
 	int tid = kernel->tids[cpu];
 	struct kbuffer *kbuf = kernel->kbufs[cpu];
 	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
 
+	/* check dummy 'time extend' record at the beginning */
+	if (raw->kbuf_offset == 0x18) {
+		uint64_t offset = 0x10;
+		unsigned long long timestamp = 0;
+		void *data = kbuffer_read_at_offset(kbuf, offset, NULL);
+		unsigned char *tmp = data - 12;  /* data still returns next record */
+
+		if ((*tmp & 0x1f) == KBUFFER_TYPE_TIME_EXTEND) {
+			uint32_t upper, lower;
+			int size;
+
+			size = kbuffer_event_size(kbuf);
+
+			memcpy(&lower, tmp, 4);
+			memcpy(&upper, tmp + 4, 4);
+			timestamp = ((uint64_t)upper << 27) + (lower >> 5);
+
+			pr_time(frs->time - timestamp);
+			pr_out("%5d: [%s] %s (+%"PRIu64" nsec)\n",
+			       tid, "time ", "extend", timestamp);
+
+			if (debug)
+				pr_hex(&offset, tmp, 8);
+			else if (kbuffer_next_event(kbuf, NULL))
+				raw->kbuf_offset += size + 4;  // 4 = event header size
+			else
+				raw->kbuf_offset = 0;
+		}
+	}
+
 	pr_time(frs->time);
 	pr_out("%5d: [%s] %s(%"PRIx64") depth: %u\n",
-	       tid, frs->type == FTRACE_EXIT ? "exit " :
-	       frs->type == FTRACE_ENTRY ? "entry" : "lost",
+	       tid, rstack_type(frs),
 	       name, frs->addr, frs->depth);
 
 	if (debug) {
@@ -475,7 +661,43 @@ static void print_raw_kernel_rstack(struct uftrace_dump_ops *ops,
 
 		size = kbuffer_event_size(kbuf);
 		raw->file_offset = kernel->offsets[cpu] + kbuffer_curr_offset(kbuf);
-		pr_hex(&raw->file_offset, data, size);
+		pr_hex(&raw->file_offset, data - 4, size + 4);
+
+		if (kbuffer_next_event(kbuf, NULL))
+			raw->kbuf_offset += size + 4;  // 4 = event header size
+		else
+			raw->kbuf_offset = 0;
+	}
+}
+
+
+static void print_raw_kernel_event(struct uftrace_dump_ops *ops,
+				   struct uftrace_kernel_reader *kernel, int cpu,
+				   struct uftrace_record *frs)
+{
+	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
+	struct event_format *event;
+	int tid = kernel->tids[cpu];
+	char *event_data;
+	int size = 0;
+
+	event = pevent_find_event(kernel->pevent, frs->addr);
+	event_data = read_kernel_event(kernel, cpu, &size);
+
+	pr_time(frs->time);
+	pr_out("%5d: [%s] %s:%s(%d) %.*s\n",
+	       tid, rstack_type(frs), event->system, event->name,
+	       frs->addr, size, event_data);
+
+	if (debug) {
+		/* this is only needed for hex dump */
+		struct kbuffer *kbuf = kernel->kbufs[cpu];
+		void *data = kbuffer_read_at_offset(kbuf, raw->kbuf_offset, NULL);
+		int size;
+
+		size = kbuffer_event_size(kbuf);
+		raw->file_offset = kernel->offsets[cpu] + kbuffer_curr_offset(kbuf);
+		pr_hex(&raw->file_offset, data - 4, size + 4);
 
 		if (kbuffer_next_event(kbuf, NULL))
 			raw->kbuf_offset += size + 4;  // 4 = event header size
@@ -491,6 +713,36 @@ static void print_raw_kernel_lost(struct uftrace_dump_ops *ops,
 	pr_red("%5d: [%s ]: %d events\n", tid, "lost", losts);
 }
 
+static void print_raw_perf_start(struct uftrace_dump_ops *ops,
+				 struct uftrace_perf_reader *perf, int cpu)
+{
+	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
+
+	if (cpu == 0)
+		pr_out("\n");
+
+	pr_out("reading perf-cpu%d.dat\n", cpu);
+
+	raw->file_offset = 0;
+}
+
+static void print_raw_perf_event(struct uftrace_dump_ops *ops,
+				 struct uftrace_perf_reader *perf,
+				 struct uftrace_record *frs)
+{
+	struct uftrace_raw_dump *raw = container_of(ops, typeof(*raw), ops);
+	char *evt_name = get_event_name(NULL, frs->addr);
+
+	pr_time(frs->time);
+	pr_out("%5d: [%s] %s(%d)\n",
+	       perf->ctxsw.tid, rstack_type(frs), evt_name, frs->addr);
+
+	if (debug)
+		pr_hex(&raw->file_offset, &perf->ctxsw, sizeof(perf->ctxsw));
+
+	free(evt_name);
+}
+
 static void print_raw_footer(struct uftrace_dump_ops *ops,
 			     struct ftrace_file_handle *handle,
 			     struct opts *opts)
@@ -523,18 +775,29 @@ static void print_chrome_task_rstack(struct uftrace_dump_ops *ops,
 {
 	char ph;
 	char spec_buf[1024];
-	struct ftrace_ret_stack *frs = task->rstack;
-	enum argspec_string_bits str_mode = NEEDS_ESCAPE;
+	struct uftrace_record *frs = task->rstack;
+	enum argspec_string_bits str_mode = NEEDS_ESCAPE | NEEDS_PAREN;
 	struct uftrace_chrome_dump *chrome = container_of(ops, typeof(*chrome), ops);
 
+	if (frs->type == UFTRACE_EVENT) {
+		if (frs->addr != EVENT_ID_PERF_SCHED_IN &&
+		    frs->addr != EVENT_ID_PERF_SCHED_OUT)
+			return;
+
+		/* new thread starts with sched-in event which should be ignored */
+		if (frs->addr == EVENT_ID_PERF_SCHED_IN && task->timestamp_last == 0)
+			return;
+	}
+
 	if (chrome->last_comma)
 		pr_out(",\n");
 	chrome->last_comma = true;
 
-	if (frs->type == FTRACE_ENTRY) {
+	if ((frs->type == UFTRACE_ENTRY) ||
+	    (frs->type == UFTRACE_EVENT && frs->addr == EVENT_ID_PERF_SCHED_OUT)) {
 		ph = 'B';
 		pr_out("{\"ts\":%"PRIu64".%03d,\"ph\":\"%c\",\"pid\":%d,\"name\":\"%s\"",
-			frs->time / 1000, frs->time % 1000, ph, task->tid, name);
+		       frs->time / 1000, (int)(frs->time % 1000), ph, task->tid, name);
 		if (frs->more) {
 			str_mode |= HAS_MORE;
 			get_argspec_string(task, spec_buf, sizeof(spec_buf), str_mode);
@@ -544,10 +807,11 @@ static void print_chrome_task_rstack(struct uftrace_dump_ops *ops,
 		else
 			pr_out("}");
 	}
-	else if (frs->type == FTRACE_EXIT) {
+	else if ((frs->type == UFTRACE_EXIT) ||
+		 (frs->type == UFTRACE_EVENT && frs->addr == EVENT_ID_PERF_SCHED_IN)) {
 		ph = 'E';
 		pr_out("{\"ts\":%"PRIu64".%03d,\"ph\":\"%c\",\"pid\":%d,\"name\":\"%s\"",
-			frs->time / 1000, frs->time % 1000, ph, task->tid, name);
+		       frs->time / 1000, (int)(frs->time % 1000), ph, task->tid, name);
 		if (frs->more) {
 			str_mode |= IS_RETVAL | HAS_MORE;
 			get_argspec_string(task, spec_buf, sizeof(spec_buf), str_mode);
@@ -557,23 +821,29 @@ static void print_chrome_task_rstack(struct uftrace_dump_ops *ops,
 		else
 			pr_out("}");
 	}
-	else
+	else if (frs->type == UFTRACE_LOST)
 		chrome->lost_event_cnt++;
 }
 
 static void print_chrome_kernel_start(struct uftrace_dump_ops *ops,
-				      struct ftrace_kernel *kernel)
+				      struct uftrace_kernel_reader *kernel)
 {
 }
 
 static void print_chrome_cpu_start(struct uftrace_dump_ops *ops,
-				   struct ftrace_kernel *kernel, int cpu)
+				   struct uftrace_kernel_reader *kernel, int cpu)
 {
 }
 
 static void print_chrome_kernel_rstack(struct uftrace_dump_ops *ops,
-				       struct ftrace_kernel *kernel, int cpu,
-				       struct ftrace_ret_stack *frs, char *name)
+				       struct uftrace_kernel_reader *kernel, int cpu,
+				       struct uftrace_record *frs, char *name)
+{
+}
+
+static void print_chrome_kernel_event(struct uftrace_dump_ops *ops,
+				      struct uftrace_kernel_reader *kernel, int cpu,
+				      struct uftrace_record *frs)
 {
 }
 
@@ -582,6 +852,17 @@ static void print_chrome_kernel_lost(struct uftrace_dump_ops *ops,
 {
 }
 
+static void print_chrome_perf_start(struct uftrace_dump_ops *ops,
+				    struct uftrace_perf_reader *perf, int cpu)
+{
+}
+
+static void print_chrome_perf_event(struct uftrace_dump_ops *ops,
+				    struct uftrace_perf_reader *perf,
+				    struct uftrace_record *frs)
+{
+}
+
 static void print_chrome_footer(struct uftrace_dump_ops *ops,
 				struct ftrace_file_handle *handle,
 				struct opts *opts)
@@ -801,15 +1082,24 @@ static void print_flame_inverted_time(struct uftrace_dump_ops *ops,
 static void print_flame_task_rstack(struct uftrace_dump_ops *ops,
 				    struct ftrace_task_handle *task, char *name)
 {
-	struct ftrace_ret_stack *frs = task->rstack;
+	struct uftrace_record *frs = task->rstack;
 	struct uftrace_flame_dump *flame = container_of(ops, typeof(*flame), ops);
 	struct fg_task *t = find_fg_task(&flame->tasks, task->tid);
 	struct fg_node *node = t->node;
 
-	if (frs->type == FTRACE_ENTRY)
+	if (frs->type == UFTRACE_ENTRY)
 		node = add_fg_node(node, name);
-	else if (frs->type == FTRACE_EXIT)
+	else if (frs->type == UFTRACE_EXIT)
 		node = add_fg_time(node, task, flame->sample_time);
+	else if (frs->type == UFTRACE_EVENT) {
+		if (frs->addr == EVENT_ID_PERF_SCHED_OUT)
+			node = add_fg_node(node, name);
+		else if (frs->addr == EVENT_ID_PERF_SCHED_IN &&
+			 task->timestamp_last) /* ignore first sched-in for thread */
+			node = add_fg_time(node, task, flame->sample_time);
+		else
+			node = t->node;  /* skip other event records */
+	}
 	else
 		node = &fg_root;
 
@@ -820,18 +1110,24 @@ static void print_flame_task_rstack(struct uftrace_dump_ops *ops,
 }
 
 static void print_flame_kernel_start(struct uftrace_dump_ops *ops,
-				     struct ftrace_kernel *kernel)
+				     struct uftrace_kernel_reader *kernel)
 {
 }
 
 static void print_flame_cpu_start(struct uftrace_dump_ops *ops,
-				  struct ftrace_kernel *kernel, int cpu)
+				  struct uftrace_kernel_reader *kernel, int cpu)
 {
 }
 
 static void print_flame_kernel_rstack(struct uftrace_dump_ops *ops,
-				      struct ftrace_kernel *kernel, int cpu,
-				      struct ftrace_ret_stack *frs, char *name)
+				      struct uftrace_kernel_reader *kernel, int cpu,
+				      struct uftrace_record *frs, char *name)
+{
+}
+
+static void print_flame_kernel_event(struct uftrace_dump_ops *ops,
+				     struct uftrace_kernel_reader *kernel, int cpu,
+				     struct uftrace_record *frs)
 {
 }
 
@@ -840,6 +1136,17 @@ static void print_flame_kernel_lost(struct uftrace_dump_ops *ops,
 {
 }
 
+static void print_flame_perf_start(struct uftrace_dump_ops *ops,
+				   struct uftrace_perf_reader *perf, int cpu)
+{
+}
+
+static void print_flame_perf_event(struct uftrace_dump_ops *ops,
+				   struct uftrace_perf_reader *perf,
+				   struct uftrace_record *frs)
+{
+}
+
 static void print_flame_footer(struct uftrace_dump_ops *ops,
 			       struct ftrace_file_handle *handle,
 			       struct opts *opts)
@@ -853,17 +1160,15 @@ static void do_dump_file(struct uftrace_dump_ops *ops, struct opts *opts,
 	int i;
 	uint64_t prev_time;
 	struct ftrace_task_handle *task;
+	struct uftrace_session_link *sessions = &handle->sessions;
 
 	ops->header(ops, handle, opts);
 
 	for (i = 0; i < handle->info.nr_tid; i++) {
-		int tid;
-
 		if (opts->kernel && opts->kernel_only)
 			continue;
 
 		task = &handle->tasks[i];
-		tid = task->tid;
 		task->rstack = &task->ustack;
 
 		prev_time = 0;
@@ -871,12 +1176,15 @@ static void do_dump_file(struct uftrace_dump_ops *ops, struct opts *opts,
 		ops->task_start(ops, task);
 
 		while (!read_task_ustack(handle, task) && !uftrace_done) {
-			struct ftrace_ret_stack *frs = &task->ustack;
-			struct ftrace_session *sess = find_task_session(tid, frs->time);
-			struct symtabs *symtabs;
-			struct sym *sym = NULL;
+			struct uftrace_record *frs = &task->ustack;
+			struct sym *sym;
 			char *name;
 
+			if (frs->more) {
+				add_to_rstack_list(&task->rstack_list,
+						   frs, &task->args);
+			}
+
 			/* consume the rstack as it didn't call read_rstack() */
 			fstack_consume(handle, task);
 
@@ -890,15 +1198,7 @@ static void do_dump_file(struct uftrace_dump_ops *ops, struct opts *opts,
 			if (!fstack_check_filter(task))
 				continue;
 
-			if (sess) {
-				symtabs = &sess->symtabs;
-				sym = find_symtabs(symtabs, frs->addr);
-
-				if (sym == NULL)
-					sym = session_find_dlsym(sess,
-								 frs->time,
-								 frs->addr);
-			}
+			sym = task_find_sym(sessions, task, frs);
 
 			name = symbol_getname(sym, frs->addr);
 			ops->task_rstack(ops, task, name);
@@ -906,22 +1206,23 @@ static void do_dump_file(struct uftrace_dump_ops *ops, struct opts *opts,
 		}
 	}
 
-	if (!opts->kernel || handle->kern == NULL || uftrace_done)
-		goto footer;
+	if (!has_kernel_data(handle->kernel) || uftrace_done)
+		goto perf;
 
-	ops->kernel_start(ops, handle->kern);
+	ops->kernel_start(ops, handle->kernel);
 
-	for (i = 0; i < handle->kern->nr_cpus; i++) {
-		struct ftrace_kernel *kernel = handle->kern;
-		struct ftrace_ret_stack *frs = &kernel->rstacks[i];
-		struct sym *sym;
-		char *name;
+	for (i = 0; i < handle->kernel->nr_cpus; i++) {
+		struct uftrace_kernel_reader *kernel = handle->kernel;
+		struct uftrace_record *frs = &kernel->rstacks[i];
+		struct uftrace_session *fsess = handle->sessions.first;
 
 		ops->cpu_start(ops, kernel, i);
 
 		while (!read_kernel_cpu_data(kernel, i) && !uftrace_done) {
 			int tid = kernel->tids[i];
 			int losts = kernel->missed_events[i];
+			struct sym *sym = NULL;
+			char *name;
 
 			if (losts) {
 				ops->lost(ops, frs->time, tid, losts);
@@ -931,15 +1232,43 @@ static void do_dump_file(struct uftrace_dump_ops *ops, struct opts *opts,
 			if (!check_time_range(&handle->time_range, frs->time))
 				continue;
 
-			sym = find_symtabs(NULL, frs->addr);
+			if (frs->type == UFTRACE_EVENT) {
+				ops->kernel_event(ops, kernel, i, frs);
+				continue;
+			}
+
+			sym = find_symtabs(&fsess->symtabs, frs->addr);
 			name = symbol_getname(sym, frs->addr);
 
-			ops->kernel(ops, kernel, i, frs, name);
+			ops->kernel_func(ops, kernel, i, frs, name);
 
 			symbol_putname(sym, name);
 		}
 	}
 
+perf:
+	if (!has_perf_data(handle) || uftrace_done)
+		goto footer;
+
+	/* force get_perf_record() to read from file */
+	handle->last_perf_idx = -1;
+
+	for (i = 0; i < handle->nr_perf; i++) {
+		struct uftrace_perf_reader *perf = &handle->perf[i];
+
+		ops->perf_start(ops, perf, i);
+
+		while (!uftrace_done) {
+			struct uftrace_record *rec;
+
+			rec = get_perf_record(handle, perf);
+			if (rec == NULL)
+				break;
+
+			ops->perf_event(ops, perf, rec);
+		}
+	}
+
 footer:
 	ops->footer(ops, handle, opts);
 }
@@ -947,41 +1276,50 @@ footer:
 static bool check_task_rstack(struct ftrace_task_handle *task,
 			      struct opts *opts)
 {
-	struct ftrace_ret_stack *frs = task->rstack;
+	struct uftrace_record *frs = task->rstack;
 
-	if (opts->kernel) {
+	if (has_kernel_data(task->h->kernel)) {
 		if (opts->kernel_skip_out) {
 			if (!task->user_stack_count &&
-			    is_kernel_address(frs->addr))
+			    is_kernel_record(task, frs))
 				return false;
 		}
 
 		if (opts->kernel_only &&
-		    !is_kernel_address(frs->addr))
+		    !is_kernel_record(task, frs))
+			return false;
+	}
+
+	if (frs->type == UFTRACE_EVENT) {
+		if (!task->user_stack_count && opts->event_skip_out)
 			return false;
 	}
 
 	if (!fstack_check_filter(task))
 		return false;
 
+	if (!check_time_range(&task->h->time_range, frs->time))
+		return false;
+
 	return true;
 }
 
 static void dump_replay_task(struct uftrace_dump_ops *ops,
 			     struct ftrace_task_handle *task)
 {
-	struct ftrace_ret_stack *frs = task->rstack;
-	struct ftrace_session *sess;
+	struct uftrace_record *frs = task->rstack;
+	struct uftrace_session_link *sessions = &task->h->sessions;
 	struct sym *sym = NULL;
 	char *name;
+	struct sym sched_sym = {
+		.name = "linux:schedule",
+	};
 
-	sess = find_task_session(task->tid, frs->time);
-	if (sess || is_kernel_address(frs->addr)) {
-		sym = find_symtabs(&sess->symtabs, frs->addr);
-		if (sym == NULL && sess)
-			sym = session_find_dlsym(sess, frs->time,
-						 frs->addr);
-	}
+	if (frs->type != UFTRACE_EVENT)
+		sym = task_find_sym(sessions, task, frs);
+	else if (frs->addr == EVENT_ID_PERF_SCHED_IN ||
+		 frs->addr == EVENT_ID_PERF_SCHED_OUT)
+		sym = &sched_sym;
 
 	name = symbol_getname(sym, frs->addr);
 	ops->task_rstack(ops, task, name);
@@ -998,7 +1336,7 @@ static void do_dump_replay(struct uftrace_dump_ops *ops, struct opts *opts,
 	ops->header(ops, handle, opts);
 
 	while (!read_rstack(handle, &task) && !uftrace_done) {
-		struct ftrace_ret_stack *frs = task->rstack;
+		struct uftrace_record *frs = task->rstack;
 
 		if (!check_task_rstack(task, opts))
 			continue;
@@ -1008,6 +1346,8 @@ static void do_dump_replay(struct uftrace_dump_ops *ops, struct opts *opts,
 		prev_time = frs->time;
 
 		dump_replay_task(ops, task);
+
+		task->timestamp_last = frs->time;
 	}
 
 	/* add duration of remaining functions */
@@ -1021,11 +1361,12 @@ static void do_dump_replay(struct uftrace_dump_ops *ops, struct opts *opts,
 
 		last_time = task->rstack->time;
 
-		if (handle->time_range.stop)
+		if (handle->time_range.stop && handle->time_range.stop < last_time)
 			last_time = handle->time_range.stop;
 
 		while (--task->stack_count >= 0) {
 			struct fstack *fstack;
+			struct uftrace_session *fsess = handle->sessions.first;
 
 			fstack = &task->func_stack[task->stack_count];
 
@@ -1042,9 +1383,14 @@ static void do_dump_replay(struct uftrace_dump_ops *ops, struct opts *opts,
 			if (task->stack_count > 0)
 				fstack[-1].child_time += fstack->total_time;
 
-			task->rstack = &task->ustack;
+			/* make sure is_kernel_record() working correctly */
+			if (is_kernel_address(&fsess->symtabs, fstack->addr))
+				task->rstack = &task->kstack;
+			else
+				task->rstack = &task->ustack;
+
 			task->rstack->time = last_time;
-			task->rstack->type = FTRACE_EXIT;
+			task->rstack->type = UFTRACE_EXIT;
 			task->rstack->addr = fstack->addr;
 
 			if (check_task_rstack(task, opts))
@@ -1059,19 +1405,11 @@ int command_dump(int argc, char *argv[], struct opts *opts)
 {
 	int ret;
 	struct ftrace_file_handle handle;
-	struct ftrace_kernel kern;
 
 	ret = open_data_file(opts, &handle);
-	if (ret < 0)
-		pr_err("cannot open data: %s", opts->dirname);
-
-	if (opts->kernel && (handle.hdr.feat_mask & KERNEL)) {
-		kern.output_dir = opts->dirname;
-		kern.skip_out = opts->kernel_skip_out;
-		if (setup_kernel_data(&kern) == 0) {
-			handle.kern = &kern;
-			load_kernel_symbol(opts->dirname);
-		}
+	if (ret < 0) {
+		pr_warn("cannot open record data: %s: %m\n", opts->dirname);
+		return -1;
 	}
 
 	fstack_setup_filters(opts, &handle);
@@ -1085,8 +1423,11 @@ int command_dump(int argc, char *argv[], struct opts *opts)
 				.task_rstack    = print_chrome_task_rstack,
 				.kernel_start   = print_chrome_kernel_start,
 				.cpu_start      = print_chrome_cpu_start,
-				.kernel         = print_chrome_kernel_rstack,
+				.kernel_func    = print_chrome_kernel_rstack,
+				.kernel_event   = print_chrome_kernel_event,
 				.lost           = print_chrome_kernel_lost,
+				.perf_start     = print_chrome_perf_start,
+				.perf_event     = print_chrome_perf_event,
 				.footer         = print_chrome_footer,
 			},
 		};
@@ -1102,8 +1443,11 @@ int command_dump(int argc, char *argv[], struct opts *opts)
 				.task_rstack    = print_flame_task_rstack,
 				.kernel_start   = print_flame_kernel_start,
 				.cpu_start      = print_flame_cpu_start,
-				.kernel         = print_flame_kernel_rstack,
+				.kernel_func    = print_flame_kernel_rstack,
+				.kernel_event   = print_flame_kernel_event,
 				.lost           = print_flame_kernel_lost,
+				.perf_start     = print_flame_perf_start,
+				.perf_event     = print_flame_perf_event,
 				.footer         = print_flame_footer,
 			},
 			.tasks = RB_ROOT,
@@ -1121,8 +1465,11 @@ int command_dump(int argc, char *argv[], struct opts *opts)
 				.task_rstack    = print_raw_task_rstack,
 				.kernel_start   = print_raw_kernel_start,
 				.cpu_start      = print_raw_cpu_start,
-				.kernel         = print_raw_kernel_rstack,
+				.kernel_func    = print_raw_kernel_rstack,
+				.kernel_event   = print_raw_kernel_event,
 				.lost           = print_raw_kernel_lost,
+				.perf_start     = print_raw_perf_start,
+				.perf_event     = print_raw_perf_event,
 				.footer         = print_raw_footer,
 			},
 		};
@@ -1130,9 +1477,6 @@ int command_dump(int argc, char *argv[], struct opts *opts)
 		do_dump_file(&dump.ops, opts, &handle);
 	}
 
-	if (handle.kern)
-		finish_kernel_data(handle.kern);
-
 	close_data_file(opts, &handle);
 
 	return ret;
