@@ -1,11 +1,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
-#include <string.h>
 #include <inttypes.h>
 #include <stdio_ext.h>
 #include <assert.h>
-#include <ctype.h>
 
 #include "uftrace.h"
 #include "utils/utils.h"
@@ -33,37 +31,58 @@ struct graph_node {
 	struct graph_node *parent;
 };
 
+enum node_type {
+	NODE_T_NORMAL,
+	NODE_T_FORK,
+	NODE_T_EXEC,
+};
+
+struct special_node {
+	struct list_head list;
+	struct graph_node *node;
+	enum node_type type;
+	int pid;
+};
+
 struct uftrace_graph {
 	char *func;
 	bool kernel_only;
-	struct ftrace_session *sess;
+	struct uftrace_session *sess;
 	struct uftrace_graph *next;
 	struct graph_backtrace *bt_curr;
 	struct list_head bt_list;
+	struct list_head special_nodes;
 	struct graph_node root;
 };
 
 struct task_graph {
 	int enabled;
 	bool lost;
+	bool new_sess;
 	struct ftrace_task_handle *task;
+	struct uftrace_graph *graph_last;
 	struct uftrace_graph *graph;
 	struct graph_node *node;
 	struct graph_backtrace *bt_curr;
 	struct rb_node link;
 };
 
+static bool full_graph = false;
 static struct rb_root tasks = RB_ROOT;
 static struct uftrace_graph *graph_list = NULL;
 
-static int create_graph(struct ftrace_session *sess, void *func)
+static int create_graph(struct uftrace_session *sess, void *func)
 {
 	struct uftrace_graph *graph = xcalloc(1, sizeof(*graph));
 
+	pr_dbg("create graph for session %.*s (%s)\n",
+	       SESSION_ID_LEN, sess->sid, sess->exename);
+
 	graph->sess = sess;
-	graph->func = xstrdup(func);
+	graph->func = xstrdup(full_graph ? basename(sess->exename) : func);
 	INIT_LIST_HEAD(&graph->root.head);
 	INIT_LIST_HEAD(&graph->bt_list);
+	INIT_LIST_HEAD(&graph->special_nodes);
 
 	graph->next = graph_list;
 	graph_list = graph;
@@ -71,11 +90,12 @@ static int create_graph(struct ftrace_session *sess, void *func)
 	return 0;
 }
 
-static void setup_graph_list(struct opts *opts, char *func)
+static void setup_graph_list(struct ftrace_file_handle *handle, struct opts *opts,
+			     char *func)
 {
 	struct uftrace_graph *graph;
 
-	walk_sessions(create_graph, func);
+	walk_sessions(&handle->sessions, create_graph, func);
 
 	graph = graph_list;
 	while (graph) {
@@ -88,12 +108,18 @@ static struct uftrace_graph * get_graph(struct ftrace_task_handle *task,
 					uint64_t time, uint64_t addr)
 {
 	struct uftrace_graph *graph;
-	struct ftrace_session *sess;
+	struct uftrace_session_link *sessions = &task->h->sessions;
+	struct uftrace_session *sess;
+
+	sess = find_task_session(sessions, task->tid, time);
+	if (sess == NULL)
+		sess = find_task_session(sessions, task->t->pid, time);
 
-	sess = find_task_session(task->tid, time);
 	if (sess == NULL) {
-		if (is_kernel_address(addr))
-			sess = first_session;
+		struct uftrace_session *fsess = sessions->first;
+
+		if (is_kernel_address(&fsess->symtabs, addr))
+			sess = fsess;
 		else
 			return NULL;
 	}
@@ -108,12 +134,15 @@ static struct uftrace_graph * get_graph(struct ftrace_task_handle *task,
 	return NULL;
 }
 
+static int start_graph(struct task_graph *tg);
+
 static struct task_graph * get_task_graph(struct ftrace_task_handle *task,
 					  uint64_t time, uint64_t addr)
 {
 	struct rb_node *parent = NULL;
 	struct rb_node **p = &tasks.rb_node;
 	struct task_graph *tg;
+	bool new_graph = false;
 
 	while (*p) {
 		parent = *p;
@@ -128,17 +157,27 @@ static struct task_graph * get_task_graph(struct ftrace_task_handle *task,
 			p = &parent->rb_right;
 	}
 
-	tg = xmalloc(sizeof(*tg));
+	tg = xzalloc(sizeof(*tg));
 	tg->task = task;
-	tg->enabled = 0;
-	tg->lost = false;
-	tg->bt_curr = NULL;
+
+	new_graph = true;
 
 	rb_link_node(&tg->link, parent, p);
 	rb_insert_color(&tg->link, &tasks);
 
 out:
 	tg->graph = get_graph(task, time, addr);
+
+	if (tg->graph_last && tg->graph != tg->graph_last) {
+		pr_dbg("detect new session: %.*s\n",
+		       SESSION_ID_LEN, tg->graph->sess->sid);
+		tg->new_sess = true;
+	}
+	tg->graph_last = tg->graph;
+
+	if (full_graph && new_graph)
+		start_graph(tg);
+
 	return tg;
 }
 
@@ -226,7 +265,7 @@ static int start_graph(struct task_graph *tg)
 	if (!tg->enabled++) {
 		save_backtrace_addr(tg);
 
-		pr_dbg("start graph\n");
+		pr_dbg("start graph for task %d\n", tg->task->tid);
 
 		tg->node = &tg->graph->root;
 		tg->node->addr = tg->task->rstack->addr;
@@ -245,7 +284,7 @@ static int end_graph(struct task_graph *tg)
 		save_backtrace_time(tg);
 		tg->lost = false;
 
-		pr_dbg("end graph\n");
+		pr_dbg("end graph for task %d\n", tg->task->tid);
 	}
 
 	return 0;
@@ -255,20 +294,28 @@ static int add_graph_entry(struct task_graph *tg)
 {
 	struct graph_node *node = NULL;
 	struct graph_node *curr = tg->node;
-	struct ftrace_ret_stack *rstack = tg->task->rstack;
-
-	if (curr == NULL)
-		return -1;
+	struct uftrace_record *rstack = tg->task->rstack;
 
 	if (tg->lost)
 		return 1;  /* ignore kernel functions after LOST */
 
+	if (tg->new_sess) {
+		curr = &tg->graph->root;
+		tg->new_sess = false;
+	}
+
+	if (curr == NULL)
+		return -1;
+
 	list_for_each_entry(node, &curr->head, list) {
 		if (node->addr == rstack->addr)
 			break;
 	}
 
 	if (list_no_entry(node, &curr->head, list)) {
+		struct uftrace_trigger tr;
+		struct uftrace_session *sess = tg->graph->sess;
+
 		node = xcalloc(1, sizeof(*node));
 
 		node->addr = rstack->addr;
@@ -277,8 +324,36 @@ static int add_graph_entry(struct task_graph *tg)
 		node->parent = curr;
 		list_add_tail(&node->list, &node->parent->head);
 		node->parent->nr_edges++;
+
+		if (uftrace_match_filter(node->addr, &sess->fixups, &tr)) {
+			struct sym *sym;
+			struct special_node *snode;
+			enum node_type type = NODE_T_NORMAL;
+
+			sym = find_symtabs(&sess->symtabs, node->addr);
+			if (sym == NULL)
+				goto out;
+
+			if (!strcmp(sym->name, "fork") ||
+			    !strcmp(sym->name, "vfork") ||
+			    !strcmp(sym->name, "daemon"))
+				type = NODE_T_FORK;
+			else if (!strncmp(sym->name, "exec", 4))
+				type = NODE_T_EXEC;
+			else
+				goto out;
+
+			snode = xmalloc(sizeof(*snode));
+			snode->node = node;
+			snode->type = type;
+			snode->pid  = tg->task->t->pid;
+
+			/* find recent one first */
+			list_add(&snode->list, &tg->graph->special_nodes);
+		}
 	}
 
+out:
 	node->nr_calls++;
 	tg->node = node;
 
@@ -294,7 +369,8 @@ static int add_graph_exit(struct task_graph *tg)
 		return -1;
 
 	if (tg->lost) {
-		if (is_kernel_address(fstack->addr))
+		if (is_kernel_address(&tg->task->h->sessions.first->symtabs,
+				      fstack->addr))
 			return 1;
 
 		/*
@@ -304,9 +380,23 @@ static int add_graph_exit(struct task_graph *tg)
 		tg->lost = false;
 	}
 
-	if (node->addr != fstack->addr)
-		pr_dbg("broken graph - addresses not match\n");
+	if (node->addr != fstack->addr) {
+		struct special_node *snode, *tmp;
 
+		list_for_each_entry_safe(snode, tmp, &tg->graph->special_nodes, list) {
+			if (snode->node->addr == tg->task->rstack->addr &&
+			    snode->type == NODE_T_FORK &&
+			    snode->pid == tg->task->t->ppid) {
+				node = snode->node;
+				list_del(&snode->list);
+				pr_dbg("recover from fork\n");
+				goto out;
+			}
+		}
+		pr_dbg("broken graph - addresses not match\n");
+	}
+
+out:
 	node->time       += fstack->total_time;
 	node->child_time += fstack->child_time;
 
@@ -315,15 +405,34 @@ static int add_graph_exit(struct task_graph *tg)
 	return 0;
 }
 
+static int add_graph_event(struct task_graph *tg)
+{
+	struct uftrace_record *rec = tg->task->rstack;
+
+	if (rec->addr == EVENT_ID_PERF_SCHED_OUT) {
+		/* to match addr with sched-in */
+		rec->addr = EVENT_ID_PERF_SCHED_IN;
+		return add_graph_entry(tg);
+	}
+	else if (rec->addr == EVENT_ID_PERF_SCHED_IN) {
+		return add_graph_exit(tg);
+	}
+
+	return -1;
+}
+
 static int add_graph(struct task_graph *tg, int type)
 {
 	pr_dbg2("add graph (enabled: %d) %s\n", tg->enabled,
-		type == FTRACE_ENTRY ? "ENTRY" : "EXIT");
+		type == UFTRACE_ENTRY ? "ENTRY" :
+		type == UFTRACE_EXIT  ? "EXIT"  : "EVENT");
 
-	if (type == FTRACE_ENTRY)
+	if (type == UFTRACE_ENTRY)
 		return add_graph_entry(tg);
-	else if (type == FTRACE_EXIT)
+	else if (type == UFTRACE_EXIT)
 		return add_graph_exit(tg);
+	else if (type == UFTRACE_EVENT)
+		return add_graph_event(tg);
 	else
 		return 0;
 }
@@ -363,15 +472,27 @@ static void print_graph_node(struct uftrace_graph *graph,
 	struct graph_node *parent = node->parent;
 	struct graph_node *child;
 	int orig_indent = indent;
+	static struct sym sched_sym = {
+		.name = "linux:schedule",
+	};
+
+	/* XXX: what if it clashes with existing function address */
+	if (node->addr == EVENT_ID_PERF_SCHED_IN)
+		sym = &sched_sym;
+	else
+		sym = find_symtabs(&graph->sess->symtabs, node->addr);
 
-	sym = find_symtabs(&graph->sess->symtabs, node->addr);
 	symname = symbol_getname(sym, node->addr);
 
 	pr_out(" ");
 	print_time_unit(node->time);
 	pr_out(" : ");
 	pr_indent(indent_mask, indent, needs_line);
-	pr_out("(%d) %s\n", node->nr_calls, symname);
+
+	if (full_graph && node == &graph->root)
+		pr_out("(%d) %s\n", 1, graph->func);
+	else
+		pr_out("(%d) %s\n", node->nr_calls, symname);
 
 	if (node->nr_edges > 1) {
 		pr_dbg2("add mask (%d) for %s\n", indent, symname);
@@ -410,20 +531,16 @@ static int print_graph(struct uftrace_graph *graph, struct opts *opts)
 	    graph->root.nr_edges == 0)
 		return 0;
 
-	pr_out("#\n");
-	pr_out("# function graph for '%s' (session: %.16s)\n",
+	pr_out("# Function Call Graph for '%s' (session: %.16s)\n",
 	       graph->func, graph->sess->sid);
-	pr_out("#\n\n");
 
-	if (!list_empty(&graph->bt_list)) {
-		pr_out("backtrace\n");
-		pr_out("=====================================\n");
+	if (!full_graph && !list_empty(&graph->bt_list)) {
+		pr_out("=============== BACKTRACE ===============\n");
 		print_backtrace(graph);
 	}
 
 	if (graph->root.time || graph->root.nr_edges) {
-		pr_out("calling functions\n");
-		pr_out("=====================================\n");
+		pr_out("========== FUNCTION CALL GRAPH ==========\n");
 		indent_mask = xcalloc(opts->max_stack, sizeof(*indent_mask));
 		print_graph_node(graph, &graph->root, indent_mask, 0,
 				 graph->root.nr_edges > 1);
@@ -433,8 +550,8 @@ static int print_graph(struct uftrace_graph *graph, struct opts *opts)
 	return 1;
 }
 
-static void build_graph_node (struct ftrace_task_handle *task, uint64_t time,
-			      uint64_t addr, int type, char *func)
+static void build_graph_node(struct ftrace_task_handle *task, uint64_t time,
+			     uint64_t addr, int type, char *func)
 {
 	struct task_graph *tg;
 	struct sym *sym = NULL;
@@ -447,50 +564,63 @@ static void build_graph_node (struct ftrace_task_handle *task, uint64_t time,
 	/* cannot find a session for this record */
 	if (tg->graph == NULL)
 		return;
+	if (type == UFTRACE_EVENT)
+		return;
+	if (full_graph)
+		return;
 
 	sym = find_symtabs(&tg->graph->sess->symtabs, addr);
 	name = symbol_getname(sym, addr);
 
 	if (!strcmp(name, func)) {
-		if (type == FTRACE_ENTRY)
+		if (type == UFTRACE_ENTRY)
 			start_graph(tg);
-		else if (type == FTRACE_EXIT)
+		else if (type == UFTRACE_EXIT)
 			end_graph(tg);
 	}
 
 	symbol_putname(sym, name);
 }
 
-static int build_graph(struct opts *opts, struct ftrace_file_handle *handle,
+static void build_graph(struct opts *opts, struct ftrace_file_handle *handle,
 		       char *func)
 {
-	int ret = 0;
 	struct ftrace_task_handle *task;
 	struct uftrace_graph *graph;
 	uint64_t prev_time = 0;
 	int i;
 
-	setup_graph_list(opts, func);
+	setup_graph_list(handle, opts, func);
 
 	while (!read_rstack(handle, &task) && !uftrace_done) {
-		struct ftrace_ret_stack *frs = task->rstack;
+		struct uftrace_record *frs = task->rstack;
 
 		/* skip user functions if --kernel-only is set */
-		if (opts->kernel_only && !is_kernel_address(frs->addr))
+		if (opts->kernel_only && !is_kernel_record(task, frs))
 			continue;
 
 		if (opts->kernel_skip_out) {
 			/* skip kernel functions outside user functions */
 			if (!task->user_stack_count &&
-			    is_kernel_address(frs->addr))
+			    is_kernel_record(task, frs))
 				continue;
 		}
 
 		if (!fstack_check_filter(task))
 			continue;
 
-		if (frs->type == FTRACE_LOST) {
+		if (frs->type == UFTRACE_EVENT) {
+			if (!task->user_stack_count && opts->event_skip_out)
+				continue;
+
+			if (frs->addr != EVENT_ID_PERF_SCHED_IN &&
+			    frs->addr != EVENT_ID_PERF_SCHED_OUT)
+				continue;
+		}
+
+		if (frs->type == UFTRACE_LOST) {
 			struct task_graph *tg;
+			struct uftrace_session *fsess;
 
 			if (opts->kernel_skip_out && !task->user_stack_count)
 				continue;
@@ -507,7 +637,7 @@ static int build_graph(struct opts *opts, struct ftrace_file_handle *handle,
 				    !(fstack->flags & FSTACK_FL_NORECORD)) {
 					build_graph_node(task, prev_time,
 							 fstack->addr,
-							 FTRACE_EXIT, func);
+							 UFTRACE_EXIT, func);
 				}
 
 				fstack_exit(task);
@@ -515,19 +645,21 @@ static int build_graph(struct opts *opts, struct ftrace_file_handle *handle,
 			}
 
 			/* force to find a session for kernel function */
+			fsess = task->h->sessions.first;
 			tg = get_task_graph(task, prev_time,
-					    (1UL << KADDR_SHIFT));
+					    fsess->symtabs.kernel_base + 1);
 			tg->lost = true;
 
-			if (tg->enabled && is_kernel_address(tg->node->addr))
+			if (tg->enabled && is_kernel_address(&fsess->symtabs,
+							     tg->node->addr))
 				pr_dbg("not returning to user after LOST\n");
 
 			continue;
 		}
 
 		if (prev_time > frs->time) {
-			pr_log("inverted time: broken data?\n");
-			return -1;
+			pr_warn("inverted time: broken data?\n");
+			return;
 		}
 		prev_time = frs->time;
 
@@ -569,23 +701,25 @@ static int build_graph(struct opts *opts, struct ftrace_file_handle *handle,
 				fstack[-1].child_time += fstack->total_time;
 
 			build_graph_node(task, last_time, fstack->addr,
-					 FTRACE_EXIT, func);
+					 UFTRACE_EXIT, func);
 		}
 	}
 
+	if (!full_graph || uftrace_done)
+		return;
+
+	/* account execution time of each graph */
 	graph = graph_list;
-	while (graph && !uftrace_done) {
-		ret += print_graph(graph, opts);
+	while (graph) {
+		struct graph_node *node;
+
+		list_for_each_entry(node, &graph->root.head, list) {
+			graph->root.time += node->time;
+			graph->root.child_time += node->time;
+		}
+
 		graph = graph->next;
 	}
-
-	if (!ret) {
-		pr_out("uftrace: cannot find graph for '%s'\n", func);
-		if (opts_has_filter(opts))
-			pr_out("\t please check your filter settings.\n");
-	}
-
-	return 0;
 }
 
 struct find_func_data {
@@ -593,7 +727,7 @@ struct find_func_data {
 	bool found;
 };
 
-static int find_func(struct ftrace_session *s, void *arg)
+static int find_func(struct uftrace_session *s, void *arg)
 {
 	struct find_func_data *data = arg;
 	struct symtabs *symtabs = &s->symtabs;
@@ -606,7 +740,9 @@ static int find_func(struct ftrace_session *s, void *arg)
 	return data->found;
 }
 
-static void synthesize_depth_trigger(struct opts *opts, char *func)
+static void synthesize_depth_trigger(struct opts *opts,
+				     struct ftrace_file_handle *handle,
+				     char *func)
 {
 	size_t old_len = opts->trigger ? strlen(opts->trigger) : 0;
 	size_t new_len = strlen(func) + 32;
@@ -614,7 +750,7 @@ static void synthesize_depth_trigger(struct opts *opts, char *func)
 		.name = func,
 	};
 
-	walk_sessions(find_func, &ffd);
+	walk_sessions(&handle->sessions, find_func, &ffd);
 
 	opts->trigger = xrealloc(opts->trigger, old_len + new_len);
 	snprintf(opts->trigger + old_len, new_len,
@@ -626,7 +762,7 @@ int command_graph(int argc, char *argv[], struct opts *opts)
 {
 	int ret;
 	struct ftrace_file_handle handle;
-	struct ftrace_kernel kern;
+	struct uftrace_graph *graph;
 	char *func;
 
 	__fsetlocking(outfp, FSETLOCKING_BYCALLER);
@@ -634,20 +770,15 @@ int command_graph(int argc, char *argv[], struct opts *opts)
 
 	if (opts->idx)
 		func = argv[opts->idx];
-	else
-		func = "main";
+	else {
+		func = "_start";
+		full_graph = true;
+	}
 
 	ret = open_data_file(opts, &handle);
-	if (ret < 0)
+	if (ret < 0) {
+		pr_warn("cannot open record data: %s: %m\n", opts->dirname);
 		return -1;
-
-	if (opts->kernel && (handle.hdr.feat_mask & KERNEL)) {
-		kern.output_dir = opts->dirname;
-		kern.skip_out = opts->kernel_skip_out;
-		if (setup_kernel_data(&kern) == 0) {
-			handle.kern = &kern;
-			load_kernel_symbol(opts->dirname);
-		}
 	}
 
 	if (opts->depth != OPT_DEPTH_DEFAULT) {
@@ -656,17 +787,26 @@ int command_graph(int argc, char *argv[], struct opts *opts)
 		 * lead to undesired result.  Set a synthetic depth
 		 * trigger to prevent the function from filtering out.
 		 */
-		synthesize_depth_trigger(opts, func);
+		synthesize_depth_trigger(opts, &handle, func);
 	}
 
 	fstack_setup_filters(opts, &handle);
 
-	ret = build_graph(opts, &handle, func);
+	build_graph(opts, &handle, func);
 
-	if (handle.kern)
-		finish_kernel_data(handle.kern);
+	graph = graph_list;
+	while (graph && !uftrace_done) {
+		ret += print_graph(graph, opts);
+		graph = graph->next;
+	}
+
+	if (!ret && !uftrace_done) {
+		pr_out("uftrace: cannot find graph for '%s'\n", func);
+		if (opts_has_filter(opts))
+			pr_out("\t please check your filter settings.\n");
+	}
 
 	close_data_file(opts, &handle);
 
-	return ret;
+	return 0;
 }
