@@ -7,6 +7,7 @@
 #include "uftrace.h"
 #include "utils/utils.h"
 #include "utils/fstack.h"
+#include "utils/kernel.h"
 #include "libmcount/mcount.h"
 
 
@@ -60,11 +61,66 @@ static void reset_live_opts(struct opts *opts)
 
 static void sigsegv_handler(int sig)
 {
-	pr_log("Segmentation fault\n");
+	pr_warn("Segmentation fault\n");
 	cleanup_tempdir();
 	raise(sig);
 }
 
+static bool can_skip_replay(struct opts *opts, int record_result)
+{
+	if (opts->nop)
+		return true;
+
+	return false;
+}
+
+static void setup_child_environ(struct opts *opts)
+{
+	char buf[4096];
+	char *old_preload, *old_libpath;
+
+	if (opts->lib_path) {
+		strcpy(buf, opts->lib_path);
+		strcat(buf, "/libmcount:");
+	} else {
+		/* to make strcat() work */
+		buf[0] = '\0';
+	}
+
+#ifdef INSTALL_LIB_PATH
+	strcat(buf, INSTALL_LIB_PATH);
+#endif
+
+	old_libpath = getenv("LD_LIBRARY_PATH");
+	if (old_libpath) {
+		size_t len = strlen(buf) + strlen(old_libpath) + 2;
+		char *libpath = xmalloc(len);
+
+		snprintf(libpath, len, "%s:%s", buf, old_libpath);
+		setenv("LD_LIBRARY_PATH", libpath, 1);
+		free(libpath);
+	}
+	else
+		setenv("LD_LIBRARY_PATH", buf, 1);
+
+	if (opts->lib_path)
+		snprintf(buf, sizeof(buf), "%s/libmcount/libmcount.so", opts->lib_path);
+	else
+		strcpy(buf, "libmcount.so");
+
+	old_preload = getenv("LD_PRELOAD");
+	if (old_preload) {
+		size_t len = strlen(buf) + strlen(old_preload) + 2;
+		char *preload = xmalloc(len);
+
+		snprintf(preload, len, "%s:%s", buf, old_preload);
+		setenv("LD_PRELOAD", preload, 1);
+		free(preload);
+	}
+	else
+		setenv("LD_PRELOAD", buf, 1);
+}
+
 int command_live(int argc, char *argv[], struct opts *opts)
 {
 	char template[32] = "/tmp/uftrace-live-XXXXXX";
@@ -89,8 +145,22 @@ int command_live(int argc, char *argv[], struct opts *opts)
 
 	opts->dirname = template;
 
+	if (opts->list_event) {
+		if (geteuid() == 0)
+			list_kernel_events();
+
+		if (fork() == 0) {
+			setup_child_environ(opts);
+			setenv("UFTRACE_LIST_EVENT", "1", 1);
+
+			execv(opts->exename, &argv[opts->idx]);
+			abort();
+		}
+		return 0;
+	}
+
 	ret = command_record(argc, argv, opts);
-	if (!opts->nop) {
+	if (!can_skip_replay(opts, ret)) {
 		int ret2;
 
 		reset_live_opts(opts);
