@@ -1,6 +1,5 @@
 #include <stdio.h>
 #include <signal.h>
-#include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdarg.h>
@@ -12,6 +11,9 @@
 #include <sys/signalfd.h>
 #include <sys/epoll.h>
 #include <linux/limits.h>
+#include <sys/stat.h>
+#include <errno.h>
+#include <sys/wait.h>
 
 #include "uftrace.h"
 #include "utils/utils.h"
@@ -60,6 +62,7 @@ static int signal_fd(struct opts *opts)
 	sigemptyset(&mask);
 	sigaddset(&mask, SIGINT);
 	sigaddset(&mask, SIGTERM);
+	sigaddset(&mask, SIGCHLD);
 
 	if (sigprocmask(SIG_BLOCK, &mask, NULL) < 0)
 		pr_err("signal block failed");
@@ -100,12 +103,12 @@ int setup_client_socket(struct opts *opts)
 	return sock;
 }
 
-void send_trace_header(int sock, char *name)
+void send_trace_dir_name(int sock, char *name)
 {
 	ssize_t len = strlen(name);
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_HDR),
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_DIR_NAME),
 		.len   = htonl(len),
 	};
 	struct iovec iov[] = {
@@ -113,7 +116,7 @@ void send_trace_header(int sock, char *name)
 		{ .iov_base = name, .iov_len = len, },
 	};
 
-	pr_dbg2("send FTRACE_MSG_SEND_HDR\n");
+	pr_dbg2("send UFTRACE_MSG_SEND_HDR\n");
 	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
 		pr_err("send header failed");
 }
@@ -121,9 +124,9 @@ void send_trace_header(int sock, char *name)
 void send_trace_data(int sock, int tid, void *data, size_t len)
 {
 	int32_t msg_tid = htonl(tid);
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_DATA),
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_DATA),
 		.len   = htonl(sizeof(msg_tid) + len),
 	};
 	struct iovec iov[] = {
@@ -132,148 +135,132 @@ void send_trace_data(int sock, int tid, void *data, size_t len)
 		{ .iov_base = data,     .iov_len = len, },
 	};
 
-	pr_dbg2("send FTRACE_MSG_SEND_DATA\n");
+	pr_dbg2("send UFTRACE_MSG_SEND_DATA\n");
 	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
 		pr_err("send data failed");
 }
 
-void send_trace_task(int sock, struct ftrace_msg *hmsg,
-		     struct ftrace_msg_task *tmsg)
+void send_trace_kernel_data(int sock, int cpu, void *data, size_t len)
 {
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_TASK),
-		.len   = htonl(sizeof(*hmsg) + sizeof(*tmsg)),
+	int32_t msg_cpu = htonl(cpu);
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_KERNEL_DATA),
+		.len   = htonl(sizeof(msg_cpu) + len),
 	};
-	struct iovec iov[] = {
-		{ .iov_base = &msg, .iov_len = sizeof(msg), },
-		{ .iov_base = hmsg, .iov_len = sizeof(*hmsg), },
-		{ .iov_base = tmsg, .iov_len = sizeof(*tmsg), },
-	};
-
-	hmsg->magic = htons(hmsg->magic);
-	hmsg->type  = htons(hmsg->type);
-	hmsg->len   = htonl(hmsg->len);
-
-	tmsg->time = htonq(tmsg->time);
-	tmsg->pid  = htonl(tmsg->pid);
-	tmsg->tid  = htonl(tmsg->tid);
-
-	pr_dbg2("send FTRACE_MSG_SEND_TASK\n");
-	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
-		pr_err("send task data failed");
-}
-
-/* namelen is 8-byte aligned length of smsg->namelen */
-void send_trace_session(int sock, struct ftrace_msg *hmsg,
-			struct ftrace_msg_sess *smsg,
-			char *exename, int namelen)
-{
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_SESSION),
-		.len   = htonl(sizeof(*hmsg) + sizeof(*smsg)),
-	};
-	struct iovec iov[] = {
-		{ .iov_base = &msg,    .iov_len = sizeof(msg), },
-		{ .iov_base = hmsg,    .iov_len = sizeof(*hmsg), },
-		{ .iov_base = smsg,    .iov_len = sizeof(*smsg), },
-		{ .iov_base = exename, .iov_len = namelen, },
-	};
-	uint64_t sid;
-	char sidbuf[sizeof(smsg->sid) + 1];
-
-	hmsg->magic = htons(hmsg->magic);
-	hmsg->type  = htons(hmsg->type);
-	hmsg->len   = htonl(hmsg->len);
-
-	smsg->task.time = htonq(smsg->task.time);
-	smsg->task.pid  = htonl(smsg->task.pid);
-	smsg->task.tid  = htonl(smsg->task.tid);
-	smsg->namelen   = htonl(smsg->namelen);
-
-	sscanf(smsg->sid, "%016"SCNx64, &sid);
-	snprintf(sidbuf, sizeof(sidbuf), "%016"PRIx64, htonq(sid));
-	memcpy(smsg->sid, sidbuf, sizeof(smsg->sid));
-
-	pr_dbg2("send FTRACE_MSG_SEND_SESSION\n");
-	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
-		pr_err("send session data failed");
-}
-
-void send_trace_map(int sock, uint64_t sid, void *map, int len)
-{
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_MAP),
-		.len   = htonl(sizeof(sid) + len),
-	};
-	uint64_t msg_sid = htonq(sid);
 	struct iovec iov[] = {
 		{ .iov_base = &msg,     .iov_len = sizeof(msg), },
-		{ .iov_base = &msg_sid, .iov_len = sizeof(msg_sid), },
-		{ .iov_base = map,      .iov_len = len, },
+		{ .iov_base = &msg_cpu, .iov_len = sizeof(msg_cpu), },
+		{ .iov_base = data,     .iov_len = len, },
 	};
 
-	pr_dbg2("send FTRACE_MSG_SEND_MAP\n");
+	pr_dbg2("send UFTRACE_MSG_SEND_KERNEL_DATA\n");
 	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
-		pr_err("send map failed");
+		pr_err("send kernel data failed");
 }
 
-void send_trace_sym(int sock, char *symfile, void *sym, int len)
+void send_trace_perf_data(int sock, int cpu, void *data, size_t len)
 {
-	int32_t namelen = strlen(symfile);
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_SYM),
-		.len   = htonl(sizeof(namelen) + namelen + len),
+	int32_t msg_cpu = htonl(cpu);
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_PERF_DATA),
+		.len   = htonl(sizeof(msg_cpu) + len),
 	};
 	struct iovec iov[] = {
+		{ .iov_base = &msg,     .iov_len = sizeof(msg), },
+		{ .iov_base = &msg_cpu, .iov_len = sizeof(msg_cpu), },
+		{ .iov_base = data,     .iov_len = len, },
+	};
+
+	pr_dbg2("send UFTRACE_MSG_SEND_PERF_DATA\n");
+	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
+		pr_err("send kernel data failed");
+}
+
+void send_trace_metadata(int sock, const char *dirname, char *filename)
+{
+	int fd;
+	void *buf;
+	size_t len;
+	char *pathname = NULL;
+	struct stat stbuf;
+	int32_t namelen = strlen(filename);
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_META_DATA),
+		.len   = sizeof(namelen) + namelen,
+	};
+	struct iovec iov[4] = {
 		{ .iov_base = &msg,     .iov_len = sizeof(msg), },
 		{ .iov_base = &namelen, .iov_len = sizeof(namelen), },
-		{ .iov_base = symfile,  .iov_len = namelen, },
-		{ .iov_base = sym,      .iov_len = len, },
+		{ .iov_base = filename, .iov_len = namelen, },
+		{ /* to be filled */ },
 	};
 
+	xasprintf(&pathname, "%s/%s", dirname, filename);
+
+	fd = open(pathname, O_RDONLY);
+	if (fd < 0)
+		pr_err("open %s failed", pathname);
+
+	if (fstat(fd, &stbuf) < 0)
+		pr_err("stat %s failed", pathname);
+
+	len = stbuf.st_size;
+	buf = xmalloc(len);
+
+	msg.len = htonl(msg.len + len);
+	iov[3].iov_base = buf;
+	iov[3].iov_len  = len;
+
+	if (read_all(fd, buf, len) < 0)
+		pr_err("map read failed");
+
 	namelen = htonl(namelen);
 
-	pr_dbg2("send FTRACE_MSG_SEND_SYM\n");
+	pr_dbg2("send UFTRACE_MSG_SEND_META_DATA: %s\n", filename);
 	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
-		pr_err("send symfile failed");
+		pr_err("send metadata failed");
+
+	free(pathname);
+	free(buf);
+	close(fd);
 }
 
-void send_trace_info(int sock, struct ftrace_file_header *hdr,
+void send_trace_info(int sock, struct uftrace_file_header *hdr,
 		     void *info, int len)
 {
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_INFO),
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_INFO),
 		.len   = htonl(sizeof(*hdr) + len),
 	};
 	struct iovec iov[] = {
-		{ .iov_base = &msg, .iov_len = sizeof(msg), },
-		{ .iov_base = hdr,  .iov_len = sizeof(*hdr), },
-		{ .iov_base = info, .iov_len = len, },
+		{ .iov_base = &msg,     .iov_len = sizeof(msg), },
+		{ .iov_base = hdr,      .iov_len = sizeof(*hdr), },
+		{ .iov_base = info,     .iov_len = len, },
 	};
 
 	hdr->version     = htonl(hdr->version);
 	hdr->header_size = htons(hdr->header_size);
 	hdr->feat_mask   = htonq(hdr->feat_mask);
 	hdr->info_mask   = htonq(hdr->info_mask);
+	hdr->max_stack   = htons(hdr->max_stack);
 
-	pr_dbg2("send FTRACE_MSG_SEND_INFO\n");
+	pr_dbg2("send UFTRACE_MSG_SEND_INFO\n");
 	if (writev_all(sock, iov, ARRAY_SIZE(iov)) < 0)
-		pr_err("send info failed");
+		pr_err("send metadata failed");
 }
 
 void send_trace_end(int sock)
 {
-	struct ftrace_msg msg = {
-		.magic = htons(FTRACE_MSG_MAGIC),
-		.type  = htons(FTRACE_MSG_SEND_END),
+	struct uftrace_msg msg = {
+		.magic = htons(UFTRACE_MSG_MAGIC),
+		.type  = htons(UFTRACE_MSG_SEND_END),
 	};
 
-	pr_dbg2("send FTRACE_MSG_SEND_END\n");
+	pr_dbg2("send UFTRACE_MSG_SEND_END\n");
 	if (write_all(sock, &msg, sizeof(msg)) < 0)
 		pr_err("send end failed");
 }
@@ -318,7 +305,7 @@ static void write_client_file(struct client_data *c, char *filename, int nr, ...
 	close(fd);
 }
 
-static void recv_trace_header(int sock, int len)
+static void recv_trace_dir_name(int sock, int len)
 {
 	char dirname[len + 1];
 	struct client_data *client;
@@ -368,120 +355,70 @@ static void recv_trace_data(int sock, int len)
 	free(filename);
 }
 
-static void recv_trace_task(int sock, int len)
+static void recv_trace_kernel_data(int sock, int len)
 {
 	struct client_data *client;
-	struct ftrace_msg msg;
-	struct ftrace_msg_task tmsg;
+	int32_t cpu;
+	char *filename = NULL;
+	void *buffer;
 
 	client = find_client(sock);
 	if (client == NULL)
 		pr_err("no client on this socket\n");
 
-	if (read_all(sock, &msg, sizeof(msg)) < 0)
-		pr_err("recv task message failed");
+	if (read_all(sock, &cpu, sizeof(cpu)) < 0)
+		pr_err("recv cpu failed");
+	cpu = ntohl(cpu);
 
-	msg.magic = htons(msg.magic);
-	msg.type  = htons(msg.type);
-	msg.len   = htonl(msg.len);
+	xasprintf(&filename, "kernel-cpu%d.dat", cpu);
 
-	if (msg.type != FTRACE_MSG_TID && msg.type != FTRACE_MSG_FORK_END)
-		pr_err("invalid task message type: %u\n", msg.type);
+	len -= sizeof(cpu);
+	buffer = xmalloc(len);
 
-	if (read_all(sock, &tmsg, sizeof(tmsg)) < 0)
-		pr_err("recv task message failed");
+	if (read_all(sock, buffer, len) < 0)
+		pr_err("recv buffer failed");
 
-	tmsg.time = htonq(tmsg.time);
-	tmsg.pid  = htonl(tmsg.pid);
-	tmsg.tid  = htonl(tmsg.tid);
+	write_client_file(client, filename, 1, buffer, len);
 
-	write_client_file(client, "task", 2, &msg, sizeof(msg),
-			  &tmsg, sizeof(tmsg));
+	free(buffer);
+	free(filename);
 }
 
-static void recv_trace_session(int sock, int len)
+static void recv_trace_perf_data(int sock, int len)
 {
 	struct client_data *client;
-	struct ftrace_msg msg;
-	struct ftrace_msg_sess smsg;
-	uint64_t sid;
-	char sidbuf[sizeof(smsg.sid) + 1];
-	char *exename;
-	int namelen;
+	int32_t cpu;
+	char *filename = NULL;
+	void *buffer;
 
 	client = find_client(sock);
 	if (client == NULL)
 		pr_err("no client on this socket\n");
 
-	if (read_all(sock, &msg, sizeof(msg)) < 0)
-		pr_err("recv session message failed");
+	if (read_all(sock, &cpu, sizeof(cpu)) < 0)
+		pr_err("recv cpu failed");
+	cpu = ntohl(cpu);
 
-	msg.magic = htons(msg.magic);
-	msg.type  = htons(msg.type);
-	msg.len   = htonl(msg.len);
+	xasprintf(&filename, "perf-cpu%d.dat", cpu);
 
-	if (msg.type != FTRACE_MSG_SESSION)
-		pr_err("invalid session message type: %u\n", msg.type);
+	len -= sizeof(cpu);
+	buffer = xmalloc(len);
 
-	if (read_all(sock, &smsg, sizeof(smsg)) < 0)
-		pr_err("recv session message failed");
+	if (read_all(sock, buffer, len) < 0)
+		pr_err("recv buffer failed");
 
-	smsg.task.time = htonq(smsg.task.time);
-	smsg.task.pid  = htonl(smsg.task.pid);
-	smsg.task.tid  = htonl(smsg.task.tid);
-	smsg.namelen   = htonl(smsg.namelen);
+	write_client_file(client, filename, 1, buffer, len);
 
-	sscanf(smsg.sid, "%016"SCNx64, &sid);
-	snprintf(sidbuf, sizeof(sidbuf), "%016"PRIx64, htonq(sid));
-	memcpy(smsg.sid, sidbuf, sizeof(smsg.sid));
-
-	namelen = ALIGN(smsg.namelen, 8);
-	exename = xmalloc(namelen);
-
-	if (read_all(sock, exename, namelen) < 0)
-		pr_err("recv exename failed");
-
-	write_client_file(client, "task", 3, &msg, sizeof(msg),
-			  &smsg, sizeof(smsg), exename, namelen);
-
-	free(exename);
+	free(buffer);
+	free(filename);
 }
 
-static void recv_trace_map(int sock, int len)
-{
-	struct client_data *client;
-	uint64_t sid;
-	char *mapname = NULL;
-	void *mapdata;
-
-	client = find_client(sock);
-	if (client == NULL)
-		pr_err("no client on this socket\n");
-
-	if (read_all(sock, &sid, sizeof(sid)) < 0)
-		pr_err("recv map session id failed");
-
-	sid = ntohq(sid);
-	xasprintf(&mapname, "sid-%016"PRIx64".map", sid);
-
-	len -= sizeof(sid);
-	mapdata = xmalloc(len);
-
-	if (read_all(sock, mapdata, len) < 0)
-		pr_err("recv map file failed");
-
-	write_client_file(client, mapname, 1, mapdata, len);
-
-	free(mapdata);
-	free(mapname);
-}
-
-static void recv_trace_sym(int sock, int len)
+static void recv_trace_metadata(int sock, int len)
 {
 	struct client_data *client;
 	int32_t namelen;
-	char *symname = NULL;
-	void *symdata;
+	char *filename = NULL;
+	void *filedata;
 
 	client = find_client(sock);
 	if (client == NULL)
@@ -491,28 +428,29 @@ static void recv_trace_sym(int sock, int len)
 		pr_err("recv symfile name length failed");
 
 	namelen = ntohl(namelen);
-	symname = xmalloc(namelen + 1);
+	filename = xmalloc(namelen + 1);
 
-	if (read_all(sock, symname, namelen) < 0)
-		pr_err("recv symfile name failed");
-	symname[namelen] = '\0';
+	if (read_all(sock, filename, namelen) < 0)
+		pr_err("recv file name failed");
+	filename[namelen] = '\0';
 
 	len -= sizeof(namelen) + namelen;
-	symdata = xmalloc(len);
+	filedata = xmalloc(len);
 
-	if (read_all(sock, symdata, len) < 0)
+	pr_dbg2("reading %s (%d bytes)\n", filename, len);
+	if (read_all(sock, filedata, len) < 0)
 		pr_err("recv symfile failed");
 
-	write_client_file(client, symname, 1, symdata, len);
+	write_client_file(client, filename, 1, filedata, len);
 
-	free(symdata);
-	free(symname);
+	free(filedata);
+	free(filename);
 }
 
 static void recv_trace_info(int sock, int len)
 {
 	struct client_data *client;
-	struct ftrace_file_header hdr;
+	struct uftrace_file_header hdr;
 	void *info;
 
 	client = find_client(sock);
@@ -526,12 +464,13 @@ static void recv_trace_info(int sock, int len)
 	hdr.header_size = ntohs(hdr.header_size);
 	hdr.feat_mask   = ntohq(hdr.feat_mask);
 	hdr.info_mask   = ntohq(hdr.info_mask);
+	hdr.max_stack   = ntohs(hdr.max_stack);
 
 	len -= sizeof(hdr);
 	info = xmalloc(len);
 
 	if (read_all(sock, info, len) < 0)
-		pr_err("recv info file failed");
+		pr_err("recv info failed");
 
 	write_client_file(client, "info", 2, &hdr, sizeof(hdr), info, len);
 
@@ -546,6 +485,8 @@ static void recv_trace_end(int sock, int efd)
 	if (client) {
 		list_del(&client->list);
 
+		pr_dbg("wrote client data to %s\n", client->dirname);
+
 		free(client->dirname);
 		free(client);
 	}
@@ -556,6 +497,20 @@ static void recv_trace_end(int sock, int efd)
 	close(sock);
 }
 
+static void execute_run_cmd(char **argv) {
+	if (!argv)
+		return;
+
+	int pid = fork();
+	if (pid < 0)
+		pr_err("cannot start child process");
+
+	if (pid == 0) {
+		execvp(argv[0], argv);
+		pr_err("Failed to execute '%s'", argv[0]);
+	}
+}
+
 static void epoll_add(int efd, int fd, unsigned event)
 {
 	struct epoll_event ev = {
@@ -575,22 +530,26 @@ static void handle_server_sock(struct epoll_event *ev, int efd)
 	int sock = ev->data.fd;
 	struct sockaddr_in addr;
 	socklen_t len = sizeof(addr);
+	char hbuf[NI_MAXHOST];
 
 	client = accept(sock, &addr, &len);
 	if (client < 0)
 		pr_err("socket accept failed");
 
+	getnameinfo((struct sockaddr *)&addr, len, hbuf, sizeof(hbuf),
+		    NULL, 0, NI_NUMERICHOST);
+
 	epoll_add(efd, client, EPOLLIN);
-	pr_log("new connection added\n");
+	pr_dbg("new connection added from %s\n", hbuf);
 }
 
-static void handle_client_sock(struct epoll_event *ev, int efd)
+static void handle_client_sock(struct epoll_event *ev, int efd, struct opts *opts)
 {
 	int sock = ev->data.fd;
-	struct ftrace_msg msg;
+	struct uftrace_msg msg;
 
 	if (ev->events & (EPOLLERR | EPOLLHUP)) {
-		pr_log("client socket closed\n");
+		pr_dbg("client socket closed\n");
 		recv_trace_end(sock, efd);
 		return;
 	}
@@ -602,41 +561,38 @@ static void handle_client_sock(struct epoll_event *ev, int efd)
 	msg.type  = ntohs(msg.type);
 	msg.len   = ntohl(msg.len);
 
-	if (msg.magic != FTRACE_MSG_MAGIC)
+	if (msg.magic != UFTRACE_MSG_MAGIC)
 		pr_err_ns("invalid message\n");
 
 	switch (msg.type) {
-	case FTRACE_MSG_SEND_HDR:
-		pr_dbg2("receive FTRACE_MSG_SEND_HDR\n");
-		recv_trace_header(sock, msg.len);
+	case UFTRACE_MSG_SEND_DIR_NAME:
+		pr_dbg2("receive UFTRACE_MSG_SEND_DIR_NAME\n");
+		recv_trace_dir_name(sock, msg.len);
 		break;
-	case FTRACE_MSG_SEND_DATA:
-		pr_dbg2("receive FTRACE_MSG_SEND_DATA\n");
+	case UFTRACE_MSG_SEND_DATA:
+		pr_dbg2("receive UFTRACE_MSG_SEND_DATA\n");
 		recv_trace_data(sock, msg.len);
 		break;
-	case FTRACE_MSG_SEND_TASK:
-		pr_dbg2("receive FTRACE_MSG_SEND_TASK\n");
-		recv_trace_task(sock, msg.len);
+	case UFTRACE_MSG_SEND_KERNEL_DATA:
+		pr_dbg2("receive UFTRACE_MSG_SEND_KERNEL_DATA\n");
+		recv_trace_kernel_data(sock, msg.len);
 		break;
-	case FTRACE_MSG_SEND_SESSION:
-		pr_dbg2("receive FTRACE_MSG_SEND_SESSION\n");
-		recv_trace_session(sock, msg.len);
+	case UFTRACE_MSG_SEND_PERF_DATA:
+		pr_dbg2("receive UFTRACE_MSG_SEND_PERF_DATA\n");
+		recv_trace_perf_data(sock, msg.len);
 		break;
-	case FTRACE_MSG_SEND_MAP:
-		pr_dbg2("receive FTRACE_MSG_SEND_MAP\n");
-		recv_trace_map(sock, msg.len);
-		break;
-	case FTRACE_MSG_SEND_SYM:
-		pr_dbg2("receive FTRACE_MSG_SEND_SYM\n");
-		recv_trace_sym(sock, msg.len);
-		break;
-	case FTRACE_MSG_SEND_INFO:
-		pr_dbg2("receive FTRACE_MSG_SEND_INFO\n");
+	case UFTRACE_MSG_SEND_INFO:
+		pr_dbg2("receive UFTRACE_MSG_SEND_INFO\n");
 		recv_trace_info(sock, msg.len);
 		break;
-	case FTRACE_MSG_SEND_END:
-		pr_dbg2("receive FTRACE_MSG_SEND_END\n");
+	case UFTRACE_MSG_SEND_META_DATA:
+		pr_dbg2("receive UFTRACE_MSG_SEND_META_DATA\n");
+		recv_trace_metadata(sock, msg.len);
+		break;
+	case UFTRACE_MSG_SEND_END:
+		pr_dbg2("receive UFTRACE_MSG_SEND_END\n");
 		recv_trace_end(sock, efd);
+		execute_run_cmd(opts->run_cmd);
 		break;
 	default:
 		pr_dbg("unknown message: %d\n", msg.type);
@@ -646,10 +602,21 @@ static void handle_client_sock(struct epoll_event *ev, int efd)
 
 int command_recv(int argc, char *argv[], struct opts *opts)
 {
+	struct signalfd_siginfo si;
 	int sock;
 	int sigfd;
 	int efd;
 
+	if (strcmp(opts->dirname, UFTRACE_DIR_NAME)) {
+		char *dirname = "current";
+
+		if ((mkdir(opts->dirname, 0755) == 0 || errno == EEXIST) &&
+		    chdir(opts->dirname) == 0)
+			dirname = opts->dirname;
+
+		pr_dbg("saving to %s directory\n", dirname);
+	}
+
 	sock = server_socket(opts);
 	sigfd = signal_fd(opts);
 
@@ -669,12 +636,17 @@ int command_recv(int argc, char *argv[], struct opts *opts)
 			pr_err("epoll wait failed");
 
 		for (i = 0; i < len; i++) {
-			if (ev[i].data.fd == sigfd)
-				uftrace_done = true;
+			if (ev[i].data.fd == sigfd) {
+				int nr = read(sigfd, &si, sizeof si);
+				if (nr > 0 && si.ssi_signo == SIGCHLD)
+					waitpid(-1, NULL, WNOHANG);
+				else
+					uftrace_done = true;
+			}
 			else if (ev[i].data.fd == sock)
 				handle_server_sock(&ev[i], efd);
 			else
-				handle_client_sock(&ev[i], efd);
+				handle_client_sock(&ev[i], efd, opts);
 		}
 	}
 
