@@ -1,5 +1,4 @@
 #include <stdio.h>
-#include <string.h>
 #include <inttypes.h>
 #include <assert.h>
 
@@ -15,6 +14,7 @@ enum {
 	AVG_NONE,
 	AVG_TOTAL,
 	AVG_SELF,
+	AVG_ANY,
 } avg_mode = AVG_NONE;
 
 struct trace_entry {
@@ -32,12 +32,28 @@ struct trace_entry {
 	struct rb_node link;
 };
 
+/* this will be used when pair entry wasn't found for diff */
+static struct trace_entry dummy_entry;
+
+/* show percentage rather than value of diff */
+static bool diff_percent = false;
+
+/* calculate diff using absolute values */
+static bool diff_absolute = true;
+
+/* show original data as well as difference */
+static bool diff_full = false;
+
+/* maximum length of symbol */
+static int maxlen = 20;
+
 static void insert_entry(struct rb_root *root, struct trace_entry *te, bool thread)
 {
 	struct trace_entry *entry;
 	struct rb_node *parent = NULL;
 	struct rb_node **p = &root->rb_node;
 	uint64_t entry_time = 0;
+	int len = 0;
 
 	pr_dbg3("%s: [%5d] %"PRIu64"/%"PRIu64" (%lu) %-s\n",
 		__func__, te->pid, te->time_total, te->time_self, te->nr_called,
@@ -51,6 +67,8 @@ static void insert_entry(struct rb_root *root, struct trace_entry *te, bool thre
 
 		if (thread)
 			cmp = te->pid - entry->pid;
+		else if (te->sym && entry->sym)
+			cmp = strcmp(te->sym->name, entry->sym->name);
 		else
 			cmp = te->addr - entry->addr;
 
@@ -71,9 +89,15 @@ static void insert_entry(struct rb_root *root, struct trace_entry *te, bool thre
 
 			entry->time_recursive += te->time_recursive;
 
-			if (entry->sym == NULL && te->sym)
+			if (entry->sym == NULL && te->sym) {
 				entry->sym = te->sym;
 
+				if (entry->sym)
+					len = strlen(entry->sym->name);
+				if (maxlen < len)
+					maxlen = len;
+			}
+
 			return;
 		}
 
@@ -101,36 +125,21 @@ static void insert_entry(struct rb_root *root, struct trace_entry *te, bool thre
 	entry->time_max = entry_time;
 	entry->time_recursive = te->time_recursive;
 
+	if (entry->sym)
+		len = strlen(entry->sym->name);
+	if (maxlen < len)
+		maxlen = len;
+
 	rb_link_node(&entry->link, parent, p);
 	rb_insert_color(&entry->link, root);
 }
 
-static bool fill_entry(struct trace_entry *te, struct ftrace_task_handle *task,
-		       uint64_t time, uint64_t addr, struct opts *opts)
+static void fill_entry_sym(struct trace_entry *te,
+			   struct ftrace_task_handle *task,
+			   struct sym *sym, uint64_t addr)
 {
-	struct ftrace_session *sess;
-	struct sym *sym;
-	struct fstack *fstack;
 	int i;
-
-	/* skip user functions if --kernel-only is set */
-	if (opts->kernel_only && !is_kernel_address(addr))
-		return false;
-
-	if (opts->kernel_skip_out) {
-		/* skip kernel functions outside user functions */
-		if (is_kernel_address(task->func_stack[0].addr) &&
-		    is_kernel_address(addr))
-			return false;
-	}
-
-	sess = find_task_session(task->tid, time);
-	if (sess == NULL && !is_kernel_address(addr))
-		return false;
-
-	sym = find_symtabs(&sess->symtabs, addr);
-	if (sym == NULL)
-		sym = session_find_dlsym(sess, time, addr);
+	struct fstack *fstack;
 
 	fstack = &task->func_stack[task->stack_count];
 
@@ -152,7 +161,31 @@ static bool fill_entry(struct trace_entry *te, struct ftrace_task_handle *task,
 			break;
 		}
 	}
+}
 
+static bool fill_entry(struct trace_entry *te, struct ftrace_task_handle *task,
+		       uint64_t time, uint64_t addr, struct opts *opts)
+{
+	struct uftrace_session_link *sessions = &task->h->sessions;
+	struct uftrace_session *sess;
+	struct sym *sym;
+
+	sess = sessions->first;
+
+	/* skip user functions if --kernel-only is set */
+	if (opts->kernel_only && !is_kernel_address(&sess->symtabs, addr))
+		return false;
+
+	if (opts->kernel_skip_out) {
+		/* skip kernel functions outside user functions */
+		if (task->user_stack_count == 0 &&
+		    is_kernel_address(&sess->symtabs, addr))
+			return false;
+	}
+
+	sym = task_find_sym_addr(sessions, task, time, addr);
+
+	fill_entry_sym(te, task, sym, addr);
 	return true;
 }
 
@@ -160,7 +193,7 @@ static void build_function_tree(struct ftrace_file_handle *handle,
 				struct rb_root *root, struct opts *opts)
 {
 	struct trace_entry te;
-	struct ftrace_ret_stack *rstack;
+	struct uftrace_record *rstack;
 	struct ftrace_task_handle *task;
 	struct fstack *fstack;
 	int i;
@@ -168,16 +201,35 @@ static void build_function_tree(struct ftrace_file_handle *handle,
 	while (read_rstack(handle, &task) >= 0 && !uftrace_done) {
 		rstack = task->rstack;
 
-		if (rstack->type != FTRACE_LOST)
+		if (rstack->type != UFTRACE_LOST)
 			task->timestamp_last = rstack->time;
 
 		if (!fstack_check_filter(task))
 			continue;
 
-		if (rstack->type == FTRACE_ENTRY)
+		if (rstack->type == UFTRACE_ENTRY)
 			continue;
 
-		if (rstack->type == FTRACE_LOST) {
+		if (rstack->type == UFTRACE_EVENT) {
+			if (!task->user_stack_count && opts->event_skip_out)
+				continue;
+
+			if (rstack->addr == EVENT_ID_PERF_SCHED_IN) {
+				static struct sym sched_sym = {
+					.addr = EVENT_ID_PERF_SCHED_IN,
+					.size = 1,
+					.type = ST_LOCAL,
+					.name = "linux:schedule",
+				};
+
+				fill_entry_sym(&te, task, &sched_sym,
+					       sched_sym.addr);
+				insert_entry(root, &te, false);
+			}
+			continue;
+		}
+
+		if (rstack->type == UFTRACE_LOST) {
 			/* add partial duration of functions before LOST */
 			while (task->stack_count >= task->user_stack_count) {
 				fstack = &task->func_stack[task->stack_count];
@@ -195,7 +247,7 @@ static void build_function_tree(struct ftrace_file_handle *handle,
 			continue;
 		}
 
-		/* rstack->type == FTRACE_EXIT */
+		/* rstack->type == UFTRACE_EXIT */
 		if (fill_entry(&te, task, rstack->time, rstack->addr, opts))
 			insert_entry(root, &te, false);
 	}
@@ -241,13 +293,14 @@ static void build_function_tree(struct ftrace_file_handle *handle,
 
 struct sort_item {
 	const char *name;
-	int (*cmp)(struct trace_entry *a, struct trace_entry *b);
+	int (*cmp)(struct trace_entry *a, struct trace_entry *b, int column);
 	int avg_mode;
 	struct list_head list;
 };
 
 #define SORT_ITEM_BASE(_name, _field, _mode)				\
-static int cmp_##_field(struct trace_entry *a, struct trace_entry *b) 	\
+static int cmp_##_field(struct trace_entry *a, struct trace_entry *b,	\
+		int sort_column)					\
 {									\
 	if (a->_field == b->_field)					\
 		return 0;						\
@@ -262,13 +315,42 @@ static struct sort_item sort_##_field = {				\
 
 #define SORT_ITEM_DIFF(_name, _field, _mode)				\
 static int cmp_diff_##_field(struct trace_entry *a,			\
-			     struct trace_entry *b)			\
+			     struct trace_entry *b,			\
+			     int sort_column)				\
 {									\
-	double pcnt_a = 100.0 * (int64_t) a->pair->_field / a->_field;	\
-	double pcnt_b = 100.0 * (int64_t) b->pair->_field / b->_field;	\
+	double pcnt_a, pcnt_b;						\
+	int64_t diff_a, diff_b;						\
+									\
+	if (sort_column != 2) {						\
+		if (a->_field == b->_field)				\
+			return 0;					\
+		return a->_field > b->_field ? 1 : -1;			\
+	}								\
+									\
+	diff_a = a->pair->_field - a->_field;				\
+	diff_b = b->pair->_field - b->_field;				\
+									\
+	if (!diff_percent) {						\
+		if (diff_a == diff_b)					\
+			return 0;					\
+									\
+		if (diff_absolute) {					\
+			diff_a = llabs(diff_a);				\
+			diff_b = llabs(diff_b);				\
+		}							\
+		return diff_a > diff_b ? 1: -1;				\
+	}								\
+									\
+	pcnt_a = 100.0 * (int64_t) diff_a / a->_field;			\
+	pcnt_b = 100.0 * (int64_t) diff_b / b->_field;			\
 									\
 	if (pcnt_a == pcnt_b)						\
 		return 0;						\
+									\
+	if (diff_absolute) {						\
+		pcnt_a = (pcnt_a > 0) ? pcnt_a : -pcnt_a;		\
+		pcnt_b = (pcnt_b > 0) ? pcnt_b : -pcnt_b;		\
+	}								\
 	return pcnt_a > pcnt_b ? 1: -1;					\
 }									\
 static struct sort_item sort_diff_##_field = {				\
@@ -284,20 +366,43 @@ static struct sort_item sort_diff_##_field = {				\
 
 /* call count is not shown as percentage */
 static int cmp_diff_nr_called(struct trace_entry *a,
-			      struct trace_entry *b)
+			      struct trace_entry *b,
+			      int sort_column)
 {
-	long call_diff_a = a->pair->nr_called - a->nr_called;
-	long call_diff_b = b->pair->nr_called - b->nr_called;
+	long call_diff_a, call_diff_b;
+	double pcnt_a, pcnt_b;
 
-	if (call_diff_a == call_diff_b) {
-		/* call count used to same, compare original count then */
+	if (sort_column != 2) {
 		if (a->nr_called == b->nr_called)
 			return 0;
-
 		return a->nr_called > b->nr_called ? 1 : -1;
 	}
 
-	return call_diff_a > call_diff_b ? 1 : -1;
+	call_diff_a = a->pair->nr_called - a->nr_called;
+	call_diff_b = b->pair->nr_called - b->nr_called;
+
+	if (!diff_percent) {
+		if (call_diff_a == call_diff_b)
+			return 0;
+
+		if (diff_absolute) {
+			call_diff_a = llabs(call_diff_a);
+			call_diff_b = llabs(call_diff_b);
+		}
+		return call_diff_a > call_diff_b ? 1 : -1;
+	}
+
+	pcnt_a = 100.0 * call_diff_a / a->nr_called;
+	pcnt_b = 100.0 * call_diff_b / b->nr_called;
+
+	if (pcnt_a == pcnt_b)
+		return 0;
+
+	if (diff_absolute) {
+		pcnt_a = (pcnt_a > 0) ? pcnt_a : -pcnt_a;
+		pcnt_b = (pcnt_b > 0) ? pcnt_b : -pcnt_b;
+	}
+	return pcnt_a > pcnt_b ? 1 : -1;
 }
 
 static struct sort_item sort_diff_nr_called = {
@@ -308,7 +413,8 @@ static struct sort_item sort_diff_nr_called = {
 };
 
 /* exclude recursive time from total time */
-static int cmp_time_total(struct trace_entry *a, struct trace_entry *b)
+static int cmp_time_total(struct trace_entry *a, struct trace_entry *b,
+			  int sort_column)
 {
 	uint64_t a_time = a->time_total - a->time_recursive;
 	uint64_t b_time = b->time_total - b->time_recursive;
@@ -325,17 +431,46 @@ static struct sort_item sort_time_total = {
 	LIST_HEAD_INIT(sort_time_total.list)
 };
 
-static int cmp_diff_time_total(struct trace_entry *a, struct trace_entry *b)
+static int cmp_diff_time_total(struct trace_entry *a, struct trace_entry *b,
+			       int sort_column)
 {
 	uint64_t a_time = a->time_total - a->time_recursive;
 	uint64_t b_time = b->time_total - b->time_recursive;
 	uint64_t a_pair_time = a->pair->time_total - a->pair->time_recursive;
 	uint64_t b_pair_time = b->pair->time_total - b->pair->time_recursive;
-	double a_pcnt = 100.0 * a_pair_time / a_time;
-	double b_pcnt = 100.0 * b_pair_time / b_time;
+	int64_t a_diff, b_diff;
+	double a_pcnt, b_pcnt;
+
+	if (sort_column != 2) {
+		if (a_time == b_time)
+			return 0;
+		return a_time > b_time ? 1 : -1;
+	}
+
+	a_diff = a_pair_time - a_time;
+	b_diff = b_pair_time - b_time;
+
+	if (!diff_percent) {
+		if (a_diff == b_diff)
+			return 0;
+
+		if (diff_absolute) {
+			a_diff = llabs(a_diff);
+			b_diff = llabs(b_diff);
+		}
+		return a_diff > b_diff ? 1 : -1;
+	}
+
+	a_pcnt = 100.0 * a_diff / a_time;
+	b_pcnt = 100.0 * b_diff / b_time;
 
 	if (a_pcnt == b_pcnt)
 		return 0;
+
+	if (diff_absolute) {
+		a_pcnt = (a_pcnt > 0) ? a_pcnt : -a_pcnt;
+		b_pcnt = (b_pcnt > 0) ? b_pcnt : -b_pcnt;
+	}
 	return a_pcnt > b_pcnt ? 1 : -1;
 }
 
@@ -346,6 +481,26 @@ static struct sort_item sort_diff_time_total = {
 	LIST_HEAD_INIT(sort_diff_time_total.list)
 };
 
+static int cmp_func_name(struct trace_entry *a, struct trace_entry *b,
+			       int sort_column)
+{
+	return strcmp(b->sym->name, a->sym->name);
+}
+
+static struct sort_item sort_func = {
+	.name = "func",
+	.cmp = cmp_func_name,
+	.avg_mode = AVG_ANY,
+	LIST_HEAD_INIT(sort_func.list),
+};
+
+static struct sort_item sort_diff_func = {
+	.name = "func",
+	.cmp = cmp_func_name,
+	.avg_mode = AVG_ANY,
+	LIST_HEAD_INIT(sort_diff_func.list),
+};
+
 //SORT_ITEM("total", time_total, AVG_NONE);
 SORT_ITEM("self", time_self, AVG_NONE);
 SORT_ITEM_BASE("call", nr_called, AVG_NONE);
@@ -360,6 +515,7 @@ struct sort_item *all_sort_items[] = {
 	&sort_time_avg,
 	&sort_time_min,
 	&sort_time_max,
+	&sort_func,
 };
 
 struct sort_item *diff_sort_items[] = {
@@ -369,6 +525,7 @@ struct sort_item *diff_sort_items[] = {
 	&sort_diff_time_avg,
 	&sort_diff_time_min,
 	&sort_diff_time_max,
+	&sort_diff_func,
 };
 
 static LIST_HEAD(sort_list);
@@ -380,7 +537,7 @@ static int cmp_entry(struct trace_entry *a, struct trace_entry *b)
 	struct sort_item *item;
 
 	list_for_each_entry(item, &sort_list, list) {
-		ret = item->cmp(a, b);
+		ret = item->cmp(a, b, 0);
 		if (ret)
 			return ret;
 	}
@@ -419,7 +576,7 @@ static int cmp_diff_entry(struct trace_entry *a, struct trace_entry *b,
 	}
 
 	list_for_each_entry(item, sort_list_head, list) {
-		ret = item->cmp(entry_a, entry_b);
+		ret = item->cmp(entry_a, entry_b, sort_column);
 		if (ret)
 			return ret;
 	}
@@ -478,7 +635,8 @@ static void setup_sort(char *sort_keys)
 			if (strcmp(k, all_sort_items[i]->name))
 				continue;
 
-			if (all_sort_items[i]->avg_mode != (avg_mode != AVG_NONE)) {
+			if ((all_sort_items[i]->avg_mode != (avg_mode != AVG_NONE)) &&
+			    (all_sort_items[i]->avg_mode != AVG_ANY)) {
 				pr_out("uftrace: '%s' sort key %s be used with %s or %s.\n",
 				       all_sort_items[i]->name,
 				       avg_mode == AVG_NONE ? "should" : "cannot",
@@ -517,7 +675,7 @@ static void print_and_delete(struct rb_root *root,
 		entry = rb_entry(node, struct trace_entry, link);
 		print_func(entry);
 
-		if (entry->pair)
+		if (entry->pair && entry->pair != &dummy_entry)
 			free(entry->pair);
 		free(entry);
 	}
@@ -550,8 +708,8 @@ static void report_functions(struct ftrace_file_handle *handle, struct opts *opt
 {
 	struct rb_root name_tree = RB_ROOT;
 	struct rb_root sort_tree = RB_ROOT;
-	const char f_format[] = "  %10.10s  %10.10s  %10.10s  %-s\n";
-	const char line[] = "====================================";
+	const char f_format[] = "  %10.10s  %10.10s  %10.10s  %-.*s\n";
+	const char line[] = "=================================================";
 
 	build_function_tree(handle, &name_tree, opts);
 
@@ -575,24 +733,25 @@ static void report_functions(struct ftrace_file_handle *handle, struct opts *opt
 		return;
 
 	if (avg_mode == AVG_NONE)
-		pr_out(f_format, "Total time", "Self time", "Calls", "Function");
+		pr_out(f_format, "Total time", "Self time", "Calls", maxlen, "Function");
 	else if (avg_mode == AVG_TOTAL)
-		pr_out(f_format, "Avg total", "Min total", "Max total", "Function");
+		pr_out(f_format, "Avg total", "Min total", "Max total", maxlen, "Function");
 	else if (avg_mode == AVG_SELF)
-		pr_out(f_format, "Avg self", "Min self", "Max self", "Function");
+		pr_out(f_format, "Avg self", "Min self", "Max self", maxlen, "Function");
 
-	pr_out(f_format, line, line, line, line);
+	pr_out(f_format, line, line, line, maxlen, line);
 
 	print_and_delete(&sort_tree, print_function);
 }
 
 static struct sym * find_task_sym(struct ftrace_file_handle *handle,
 				  struct ftrace_task_handle *task,
-				  struct ftrace_ret_stack *rstack)
+				  struct uftrace_record *rstack)
 {
 	struct sym *sym;
 	struct ftrace_task_handle *main_task = &handle->tasks[0];
-	struct ftrace_session *sess = find_task_session(task->tid, rstack->time);
+	struct uftrace_session *sess = find_task_session(&handle->sessions,
+							 task->tid, rstack->time);
 	struct symtabs *symtabs = &sess->symtabs;
 
 	if (task->func)
@@ -634,28 +793,28 @@ static void print_thread(struct trace_entry *entry)
 static void report_threads(struct ftrace_file_handle *handle, struct opts *opts)
 {
 	struct trace_entry te;
-	struct ftrace_ret_stack *rstack;
+	struct uftrace_record *rstack;
 	struct rb_root name_tree = RB_ROOT;
 	struct ftrace_task_handle *task;
 	struct fstack *fstack;
-	const char t_format[] = "  %5.5s  %10.10s  %10.10s  %-s\n";
-	const char line[] = "====================================";
+	const char t_format[] = "  %5.5s  %10.10s  %10.10s  %-.*s\n";
+	const char line[] = "=================================================";
 
 	while (read_rstack(handle, &task) >= 0 && !uftrace_done) {
 		rstack = task->rstack;
-		if (rstack->type == FTRACE_ENTRY && task->func)
+		if (rstack->type == UFTRACE_ENTRY && task->func)
 			continue;
-		if (rstack->type == FTRACE_LOST)
+		if (rstack->type == UFTRACE_LOST)
 			continue;
 
 		/* skip user functions if --kernel-only is set */
-		if (opts->kernel_only && !is_kernel_address(rstack->addr))
+		if (opts->kernel_only && !is_kernel_record(task, rstack))
 			continue;
 
 		if (opts->kernel_skip_out) {
 			/* skip kernel functions outside user functions */
-			if (is_kernel_address(task->func_stack[0].addr) &&
-			    is_kernel_address(rstack->addr))
+			if (task->user_stack_count == 0 &&
+			    is_kernel_record(task, rstack))
 				continue;
 		}
 
@@ -664,8 +823,9 @@ static void report_threads(struct ftrace_file_handle *handle, struct opts *opts)
 		te.pid = task->tid;
 		te.sym = find_task_sym(handle, task, rstack);
 		te.addr = rstack->addr;
+		te.time_recursive = 0;
 
-		if (rstack->type == FTRACE_ENTRY) {
+		if (rstack->type == UFTRACE_ENTRY) {
 			te.time_total = te.time_self = 0;
 			te.nr_called = 0;
 		}
@@ -681,8 +841,8 @@ static void report_threads(struct ftrace_file_handle *handle, struct opts *opts)
 	if (uftrace_done)
 		return;
 
-	pr_out(t_format, "TID", "Run time", "Num funcs", "Start function");
-	pr_out(t_format, line, line, line, line);
+	pr_out(t_format, "TID", "Run time", "Num funcs", maxlen, "Start function");
+	pr_out(t_format, line, line, line, maxlen, line);
 
 	print_and_delete(&name_tree, print_thread);
 }
@@ -806,11 +966,8 @@ static void sort_function_name(struct rb_root *root_in,
 }
 
 static void calculate_diff(struct rb_root *base, struct rb_root *pair,
-			   struct rb_root *diff, struct rb_root *remaining,
-			   int sort_column)
+			   struct rb_root *diff, int sort_column)
 {
-	struct rb_root tmp = RB_ROOT;
-
 	while (!RB_EMPTY_ROOT(base) && !uftrace_done) {
 		struct rb_node *node;
 		struct trace_entry *e, *p;
@@ -820,13 +977,12 @@ static void calculate_diff(struct rb_root *base, struct rb_root *pair,
 
 		e = rb_entry(node, struct trace_entry, link);
 		p = find_by_name(pair, e);
-		if (p == NULL) {
-			sort_entries(remaining, e);
-			continue;
+		if (p != NULL) {
+			rb_erase(&p->link, pair);
+			RB_CLEAR_NODE(&p->link);
 		}
-
-		rb_erase(&p->link, pair);
-		RB_CLEAR_NODE(&p->link);
+		else
+			p = &dummy_entry;
 
 		e->pair = p;
 		p->pair = e;
@@ -834,65 +990,133 @@ static void calculate_diff(struct rb_root *base, struct rb_root *pair,
 		sort_diff_entries(diff, e, sort_column);
 	}
 
-	/* sort remaining pair entries by time */
+	/* sort remaining pair entries with zero entry */
 	while (!RB_EMPTY_ROOT(pair) && !uftrace_done) {
 		struct rb_node *node;
-		struct trace_entry *entry;
+		struct trace_entry *entry, *zero;
 
 		node = rb_first(pair);
 		rb_erase(node, pair);
 
 		entry = rb_entry(node, struct trace_entry, link);
-		sort_entries(&tmp, entry);
-	}
 
-	*pair = tmp;
+		zero = xzalloc(sizeof(*zero));
+		zero->sym = entry->sym;
+		zero->addr = entry->addr;
+		zero->pair = entry;
+		entry->pair = zero;
+
+		sort_diff_entries(diff, zero, sort_column);
+	}
 }
 
-static void print_diff(struct trace_entry *entry)
+#define NODATA "-"
+
+static void print_time_or_dash(uint64_t time_nsec)
+{
+	if (time_nsec)
+		print_time_unit(time_nsec);
+	else
+		pr_out("%10s", NODATA);
+}
+
+static void print_function_diff(struct trace_entry *entry)
 {
 	char *symname = symbol_getname(entry->sym, entry->addr);
 	struct trace_entry *pair = entry->pair;
 
 	if (avg_mode == AVG_NONE) {
 		pr_out("  ");
-		print_time_unit(entry->time_total);
-		pr_out("  ");
-		print_time_unit(pair->time_total);
-		pr_out(" ");
-		print_diff_percent(entry->time_total, pair->time_total);
+
+		if (diff_full) {
+			print_time_or_dash(entry->time_total - entry->time_recursive);
+			pr_out("  ");
+			print_time_or_dash(pair->time_total - pair->time_recursive);
+			pr_out("  ");
+		}
+		else if (diff_percent)
+			pr_out("   ");
+
+		if (diff_percent)
+			print_diff_percent(entry->time_total - entry->time_recursive,
+					   pair->time_total - pair->time_recursive);
+		else
+			print_diff_time_unit(entry->time_total - entry->time_recursive,
+					     pair->time_total - pair->time_recursive);
 
 		pr_out("   ");
-		print_time_unit(entry->time_self);
-		pr_out("  ");
-		print_time_unit(pair->time_self);
-		pr_out(" ");
-		print_diff_percent(entry->time_self, pair->time_self);
 
-		pr_out("    %9lu  %9lu  %+9ld   %-s\n",
-		       entry->nr_called, pair->nr_called,
-		       (long)(pair->nr_called - entry->nr_called), symname);
+		if (diff_full) {
+			print_time_or_dash(entry->time_self);
+			pr_out("  ");
+			print_time_or_dash(pair->time_self);
+			pr_out("  ");
+		}
+		else if (diff_percent)
+			pr_out("   ");
+
+		if (diff_percent)
+			print_diff_percent(entry->time_self, pair->time_self);
+		else
+			print_diff_time_unit(entry->time_self, pair->time_self);
+
+		pr_out("   ");
+
+		if (diff_full)
+			pr_out(" %9lu  %9lu", entry->nr_called, pair->nr_called);
+
+		pr_out("  ");
+
+		print_diff_count(entry->nr_called, pair->nr_called);
+		pr_out("   %-s\n", symname);
 	} else {
 		pr_out("  ");
-		print_time_unit(entry->time_avg);
-		pr_out("  ");
-		print_time_unit(pair->time_avg);
-		pr_out(" ");
-		print_diff_percent(entry->time_avg, pair->time_avg);
+
+		if (diff_full) {
+			print_time_unit(entry->time_avg);
+			pr_out("  ");
+			print_time_unit(pair->time_avg);
+			pr_out("  ");
+		}
+		else if (diff_percent)
+			pr_out("   ");
+
+		if (diff_percent)
+			print_diff_percent(entry->time_avg, pair->time_avg);
+		else
+			print_diff_time_unit(entry->time_avg, pair->time_avg);
 
 		pr_out("   ");
-		print_time_unit(entry->time_min);
-		pr_out("  ");
-		print_time_unit(pair->time_min);
-		pr_out(" ");
-		print_diff_percent(entry->time_min, pair->time_min);
+
+		if (diff_full) {
+			print_time_unit(entry->time_min);
+			pr_out("  ");
+			print_time_unit(pair->time_min);
+			pr_out("  ");
+		}
+		else if (diff_percent)
+			pr_out("   ");
+
+		if (diff_percent)
+			print_diff_percent(entry->time_min, pair->time_min);
+		else
+			print_diff_time_unit(entry->time_min, pair->time_min);
 
 		pr_out("   ");
-		print_time_unit(entry->time_max);
-		pr_out("  ");
-		print_time_unit(pair->time_max);
-		pr_out(" ");
-		print_diff_percent(entry->time_max, pair->time_max);
+
+		if (diff_full) {
+			print_time_unit(entry->time_max);
+			pr_out("  ");
+			print_time_unit(pair->time_max);
+			pr_out("  ");
+		}
+		else if (diff_percent)
+			pr_out("   ");
+
+		if (diff_percent)
+			print_diff_percent(entry->time_max, pair->time_max);
+		else
+			print_diff_time_unit(entry->time_max, pair->time_max);
 
 		pr_out("   %-s\n", symname);
 	}
@@ -900,71 +1124,6 @@ static void print_diff(struct trace_entry *entry)
 	symbol_putname(entry->sym, symname);
 }
 
-#define NODATA  "-"
-static void print_remaining(struct trace_entry *entry)
-{
-	char *symname = symbol_getname(entry->sym, entry->addr);
-
-	if (avg_mode == AVG_NONE) {
-		pr_out("  ");
-		print_time_unit(entry->time_total);
-		pr_out("  %10s  %8s   ", NODATA, NODATA);
-
-		print_time_unit(entry->time_self);
-		pr_out("  %10s  %8s ",  NODATA, NODATA);
-
-		pr_out("  %10lu  %9s  %9s   %-s\n",
-		       entry->nr_called, NODATA, NODATA, symname);
-	} else {
-		pr_out("  ");
-		print_time_unit(entry->time_avg);
-		pr_out("  %10s  %8s   ", NODATA, NODATA);
-
-		print_time_unit(entry->time_min);
-		pr_out("  %10s  %8s   ", NODATA, NODATA);
-
-		print_time_unit(entry->time_max);
-		pr_out("  %10s  %8s   %-s\n",  NODATA, NODATA, symname);
-	}
-
-	symbol_putname(entry->sym, symname);
-}
-
-static void print_remaining_pair(struct trace_entry *entry)
-{
-	char *symname = symbol_getname(entry->sym, entry->addr);
-
-	if (avg_mode == AVG_NONE) {
-		pr_out("  %10s  ", NODATA);
-		print_time_unit(entry->time_total);
-		pr_out("  %8s ", NODATA);
-
-		pr_out("  %10s  ", NODATA);
-		print_time_unit(entry->time_self);
-		pr_out("  %8s ", NODATA);
-
-		pr_out("  %10s %10lu %10s   %-s\n",
-		       NODATA, entry->nr_called, NODATA, symname);
-	} else {
-
-		pr_out("  %10s  ", NODATA);
-		print_time_unit(entry->time_avg);
-		pr_out("  %8s ", NODATA);
-
-		pr_out("  %10s  ", NODATA);
-		print_time_unit(entry->time_min);
-		pr_out("  %8s ", NODATA);
-
-		pr_out("  %10s  ",  NODATA);
-		print_time_unit(entry->time_max);
-		pr_out("  %8s ", NODATA);
-
-		pr_out("  %-s\n", symname);
-	}
-
-	symbol_putname(entry->sym, symname);
-}
-
 static void report_diff(struct ftrace_file_handle *handle, struct opts *opts)
 {
 	struct opts dummy_opts = {
@@ -979,22 +1138,44 @@ static void report_diff(struct ftrace_file_handle *handle, struct opts *opts)
 	struct rb_root tmp = RB_ROOT;
 	struct rb_root name_tree = RB_ROOT;
 	struct rb_root diff_tree = RB_ROOT;
-	struct rb_root remaining = RB_ROOT;
-	const char format[] = "  %32.32s   %32.32s   %32.32s   %-s\n";
-	const char line[] = "====================================";
+	const char *formats[] = {
+		"  %35.35s   %35.35s   %32.32s   %-.*s\n",  /* diff numbers */
+		"  %32.32s   %32.32s   %32.32s   %-.*s\n",  /* diff percent */
+		"  %35.35s   %35.35s   %35.35s   %-.*s\n",  /* diff avg numbers */
+		"  %11.11s   %11.11s   %11.11s   %-.*s\n",  /* diff compact */
+	};
+	const char line[] = "=================================================";
+	const char *headers[][3] = {
+		{ "Total time (diff)", "Self time (diff)", "Calls (diff)" },
+		{ "Avg total (diff)", "Min total (diff)", "Max total (diff)" },
+		{ "Avg self (diff)", "Min self (diff)", "Max self (diff)" },
+		{ "Total time", "Self time", "Calls" },
+		{ "Avg total", "Min total", "Max total" },
+		{ "Avg self", "Min self", "Max self" },
+	};
+	int h_idx = (avg_mode == AVG_NONE) ? 0 : (avg_mode == AVG_TOTAL) ? 1 : 2;
+	int f_idx = diff_percent ? 1 : (avg_mode == AVG_NONE) ? 0 : 2;
+
+	if (!diff_full) {
+		h_idx += 3;
+		f_idx = 3;
+	}
 
 	build_function_tree(handle, &tmp, opts);
 	sort_function_name(&tmp, &name_tree);
 
-	remaining = tmp;
 	tmp = RB_ROOT;
 
-	open_data_file(&dummy_opts, &data.handle);
+	if (open_data_file(&dummy_opts, &data.handle) < 0) {
+		pr_warn("cannot open record data: %s: %m\n", opts->diff);
+		goto out;
+	}
+
 	fstack_setup_filters(&dummy_opts, &data.handle);
 	build_function_tree(&data.handle, &tmp, &dummy_opts);
 	sort_function_name(&tmp, &data.root);
 
-	calculate_diff(&name_tree, &data.root, &diff_tree, &remaining, opts->sort_column);
+	calculate_diff(&name_tree, &data.root, &diff_tree, opts->sort_column);
 
 	if (uftrace_done)
 		goto out;
@@ -1004,35 +1185,50 @@ static void report_diff(struct ftrace_file_handle *handle, struct opts *opts)
 	pr_out("#  [%d] base: %s\t(from %s)\n", 0, handle->dirname, handle->info.cmdline);
 	pr_out("#  [%d] diff: %s\t(from %s)\n", 1, opts->diff, data.handle.info.cmdline);
 	pr_out("#\n");
+	pr_out(formats[f_idx], headers[h_idx][0], headers[h_idx][1], headers[h_idx][2],
+	       maxlen, "Function");
+	pr_out(formats[f_idx], line, line, line, maxlen, line);
 
-	if (avg_mode == AVG_NONE)
-		pr_out(format, "Total time (diff)", "Self time (diff)",
-		       "Nr. called (diff)", "Function");
-	else if (avg_mode == AVG_TOTAL)
-		pr_out(format, "Avg total (diff)", "Min total (diff)",
-		       "Max total (diff)", "Function");
-	else if (avg_mode == AVG_SELF)
-		pr_out(format, "Avg self (diff)", "Min self (diff)",
-		       "Max self (diff)", "Function");
-
-	pr_out(format, line, line, line, line);
-
-	print_and_delete(&remaining, print_remaining);
-	print_and_delete(&data.root, print_remaining_pair);
-	print_and_delete(&diff_tree, print_diff);
+	print_and_delete(&diff_tree, print_function_diff);
 
 out:
 	close_data_file(&dummy_opts, &data.handle);
 }
 
+static void apply_diff_policy(char *policy)
+{
+	char *str = xstrdup(policy);
+	char *p, *tmp = policy;
+
+	while ((p = strtok(tmp, ",")) != NULL) {
+		bool on = true;
+
+		if (!strncmp(p, "no-", 3)) {
+			on = false;
+			p += 3;
+		}
+
+		if (!strncmp(p, "abs", 3))
+			diff_absolute = on;
+		else if (!strncmp(p, "percent", 7))
+			diff_percent = on;
+		else if (!strncmp(p, "full", 4))
+			diff_full = true;
+		else if (!strncmp(p, "compact", 7))
+			diff_full = false;
+
+		tmp = NULL;
+	}
+	free(str);
+}
+
 int command_report(int argc, char *argv[], struct opts *opts)
 {
 	int ret;
 	struct ftrace_file_handle handle;
-	struct ftrace_kernel kern;
 
 	if (opts->avg_total && opts->avg_self) {
-		pr_out("--avg-total and --avg-self options should not be used together.\n");
+		pr_use("--avg-total and --avg-self options should not be used together.\n");
 		exit(1);
 	} else if (opts->avg_total)
 		avg_mode = AVG_TOTAL;
@@ -1040,16 +1236,9 @@ int command_report(int argc, char *argv[], struct opts *opts)
 		avg_mode = AVG_SELF;
 
 	ret = open_data_file(opts, &handle);
-	if (ret < 0)
+	if (ret < 0) {
+		pr_warn("cannot open record data: %s: %m\n", opts->dirname);
 		return -1;
-
-	if (opts->kernel && (handle.hdr.feat_mask & KERNEL)) {
-		kern.output_dir = opts->dirname;
-		kern.skip_out = opts->kernel_skip_out;
-		if (setup_kernel_data(&kern) == 0) {
-			handle.kern = &kern;
-			load_kernel_symbol(opts->dirname);
-		}
 	}
 
 	fstack_setup_filters(opts, &handle);
@@ -1069,6 +1258,9 @@ int command_report(int argc, char *argv[], struct opts *opts)
 		}
 	}
 
+	if (opts->diff_policy)
+		apply_diff_policy(opts->diff_policy);
+
 	if (opts->report_thread)
 		report_threads(&handle, opts);
 	else if (opts->diff)
@@ -1076,9 +1268,6 @@ int command_report(int argc, char *argv[], struct opts *opts)
 	else
 		report_functions(&handle, opts);
 
-	if (handle.kern)
-		finish_kernel_data(handle.kern);
-
 	close_data_file(opts, &handle);
 
 	return ret;
