@@ -10,7 +10,7 @@ class TestCase(TestBase):
    1.776 us [18130] | getsid();
    1.289 us [18130] | getuid();
    1.043 us [18130] | getgid();
-""")
+""", sort='simple')
 
     def runcmd(self):
         return '%s -F "get.?id@plt" %s' % (TestBase.ftrace, 't-getids')
