@@ -1,6 +1,5 @@
 #!/usr/bin/env python
 
-import re
 from runtest import TestBase
 
 class TestCase(TestBase):
@@ -16,6 +15,13 @@ class TestCase(TestBase):
    2.107 ms [23157] | } /* main */
 """)
 
+    def build(self, name, cflags='', ldflags=''):
+        # cygprof doesn't support return value now
+        if cflags.find('-finstrument-functions') >= 0:
+            return TestBase.TEST_SKIP
+
+        return TestBase.build(self, name, cflags, ldflags)
+
     def runcmd(self):
         return '%s -t 1ms -R mem_alloc@retval -A mem_free@arg1 -A usleep@plt,arg1 %s' % \
             (TestBase.ftrace, 't-' + self.name)
