@@ -9,23 +9,18 @@ FUNC='main'
 class TestCase(TestBase):
     def __init__(self):
         TestBase.__init__(self, 'sort', result="""
-#
-# function graph for 'main'
-#
+# Function Call Graph for 'main' (session: baa921f86e22e0c9)
+=============== BACKTRACE ===============
+ backtrace #0: hit 1, time  11.460 ms
+   [0] main (0x40069e)
 
-backtrace
-================================
- backtrace #0: hit 1, time  10.293 ms
-   [0] main (0x4004f0)
-
-calling functions
-================================
-  10.293 ms : (1) main
-  46.626 us :  +-(2) foo
-  44.360 us :  | (6) loop
+========== FUNCTION CALL GRAPH ==========
+  11.460 ms : (1) main
+ 311.345 us :  +-(2) foo
+ 308.918 us :  | (6) loop
             :  | 
-  10.138 ms :  +-(1) bar
-  10.100 ms :    (1) usleep
+  10.362 ms :  +-(1) bar
+  10.091 ms :    (1) usleep
 """, sort='graph')
 
     def pre(self):
