@@ -1,6 +1,5 @@
 #!/usr/bin/env python
 
-import re
 from runtest import TestBase
 
 class TestCase(TestBase):
@@ -15,6 +14,13 @@ class TestCase(TestBase):
    3.281 ms [18276] | } /* main */
 """)
 
+    def build(self, name, cflags='', ldflags=''):
+        # cygprof doesn't support arguments now
+        if cflags.find('-finstrument-functions') >= 0:
+            return TestBase.TEST_SKIP
+
+        return TestBase.build(self, name, cflags, ldflags)
+
     def runcmd(self):
         argopt  = '-A "float_add@fparg1/32,fparg2/32" -R "float_add@retval/f32" '
         argopt += '-A "float_sub@fparg1/32,fparg2"    -R "float_sub@retval/f32" '
