@@ -25,8 +25,16 @@ class TestCase(TestBase):
   18.161 us [13476] | } /* main */
 """)
 
+    def build(self, name, cflags='', ldflags=''):
+        # cygprof doesn't support arguments now
+        if cflags.find('-finstrument-functions') >= 0:
+            return TestBase.TEST_SKIP
+
+        return TestBase.build(self, name, cflags, ldflags)
+
     def runcmd(self):
-        argopt  = '-A "many@arg1%stack+1,arg2%stack+2,arg3%stack+3,arg4%stack+4" '
-        argopt += '-A "many@arg5%stack5,arg6%stack6,arg7%stack7"'
+        argopt  = '-A "many@arg1/i32%stack+1,arg2/i32%stack+2" '
+        argopt += '-A "many@arg3/i32%stack+3,arg4/i32%stack+4" '
+        argopt += '-A "many@arg5/i32%stack5,arg6/i32%stack6,arg7/i32%stack7"'
 
         return '%s %s %s' % (TestBase.ftrace, argopt, 't-' + self.name)
