@@ -1,6 +1,5 @@
 #!/usr/bin/env python
 
-import re
 from runtest import TestBase
 
 class TestCase(TestBase):
@@ -16,8 +15,19 @@ class TestCase(TestBase):
   13.250 us [ 9624] | } /* main */
 """)
 
+    def build(self, name, cflags='', ldflags=''):
+        # cygprof doesn't support arguments now
+        if cflags.find('-finstrument-functions') >= 0:
+            return TestBase.TEST_SKIP
+
+        return TestBase.build(self, name, cflags, ldflags)
+
     def runcmd(self):
         argopt  = '-A "variadic@arg1/s,arg2/c,arg3/s,arg4,arg5,arg6,arg7/i64,fparg1" '
-        argopt += '-A "vsnprintf@arg2,arg3/s"'
+        argopt += '-A "vsnprintf@arg2,arg3/s" '
+        argopt += '-A "__vsnprintf_chk@arg2,arg5/s"'
 
         return '%s %s %s' % (TestBase.ftrace, argopt, 't-' + self.name)
+
+    def fixup(self, cflags, result):
+        return result.replace('vsnprintf', "__vsnprintf_chk")
