@@ -9,19 +9,14 @@ FUNC='main'
 class TestCase(TestBase):
     def __init__(self):
         TestBase.__init__(self, 'exit', result="""
-#
-# function graph for 'main' (session: 924133e5b1cd2228)
-#
+# Function Call Graph for 'main' (session: 095c3a95937bdbae)
+=============== BACKTRACE ===============
+ backtrace #0: hit 1, time   0.527 us
+   [0] main (0x400480)
 
-backtrace
-================================
- backtrace #0: hit 1, time           
-   [0] main (0x400644)
-
-calling functions
-================================
-            : (1) main
-            : (1) foo
+========== FUNCTION CALL GRAPH ==========
+   0.527 us : (1) main
+   0.387 us : (1) foo
             : (1) exit
 """, sort='graph')
 
