@@ -19,7 +19,7 @@ class TestCase(TestBase):
     4.234 us    4.234 us           1  getpid
     1.568 us    1.568 us           1  __monstartup
     1.140 us    1.140 us           1  __cxa_atexit
-""")
+""", sort='report')
 
     def pre(self):
         record_cmd = '%s record -d %s %s' % (TestBase.ftrace, TDIR, 't-' + self.name)
@@ -43,24 +43,3 @@ class TestCase(TestBase):
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
         return ret
-
-    def sort(self, output):
-        """ This function post-processes output of the test to be compared .
-            It ignores blank and comment (#) lines and remaining functions.  """
-        result = []
-        for ln in output.split('\n'):
-            if ln.strip() == '':
-                continue
-            line = ln.split()
-            if line[0] == 'Total':
-                continue
-            if line[0].startswith('='):
-                continue
-            # A report line consists of following data
-            # [0]         [1]   [2]        [3]   [4]    [5]
-            # total_time  unit  self_time  unit  calls  function
-            if line[5].startswith('__'):
-                continue
-            result.append('%s %s' % (line[4], line[5]))
-
-        return '\n'.join(result)
