@@ -9,22 +9,17 @@ FUNC='a'
 class TestCase(TestBase):
     def __init__(self):
         TestBase.__init__(self, 'fork', result="""
-#
-# function graph for 'a' (session: de8436a173b22b1c)
-#
+# Function Call Graph for 'a' (session: 5eec64959f2b2e87)
+=============== BACKTRACE ===============
+ backtrace #0: hit 1, time   4.290 us
+   [0] main (0x4005c0)
+   [1] a (0x4007a1)
 
-backtrace
-================================
- backtrace #0: hit 1, time   6.602 us
-   [0] main (0x4005c5)
-   [1] a (0x400782)
-
-calling functions
-================================
-   6.602 us : (1) a
-   6.094 us : (1) b
-   5.680 us : (1) c
-   4.234 us : (1) getpid
+========== FUNCTION CALL GRAPH ==========
+   4.290 us : (1) a
+   3.970 us : (1) b
+   3.580 us : (1) c
+   2.616 us : (1) getpid
 """, sort='graph')
 
     def pre(self):
