@@ -13,8 +13,8 @@ uftrace file header: version       = 4
 uftrace file header: header size   = 40
 uftrace file header: endian        = 1 (little)
 uftrace file header: class         = 2 (64 bit)
-uftrace file header: features      = 0x63
-uftrace file header: info          = 0x3ff
+uftrace file header: features      = 0x263 (PLTHOOK | TASK_SESSION | SYM_REL_ADDR | MAX_STACK | AUTO_ARGS)
+uftrace file header: info          = 0xbff
 
 reading 5231.dat
 58348.873430946   5231: [entry] __monstartup(4004d0) depth: 0
@@ -46,4 +46,11 @@ reading 5231.dat
         return ret
 
     def fixup(self, cflags, result):
-        return result.replace("2 (64 bit)", "1 (32 bit)")
+        import platform
+
+        if platform.architecture()[0] == '32bit':
+            result = result.replace("2 (64 bit)", "1 (32 bit)")
+        p = sp.Popen(['file', 't-' + self.name], stdout=sp.PIPE)
+        if 'BuildID' not in p.communicate()[0].decode(errors='ignore'):
+            result = result.replace("0x3ff", "0x3fd")
+        return result
