@@ -9,22 +9,16 @@ FUNC='main'
 class TestCase(TestBase):
     def __init__(self):
         TestBase.__init__(self, 'sleep', result="""
-#
-# function graph for 'main' (session: 0bc5da823389c319)
-#
-
-backtrace
-================================
- backtrace #0: hit 1, time   2.103 ms
-   [0] main (0x400550)
-
-calling functions
-================================
-   2.103 ms : (1) main
-   2.102 ms : (1) foo
-   2.084 ms : (1) bar
-   2.080 ms : (1) usleep
+# Function Call Graph for 'main' (session: b78e9c27042adaa7)
+=============== BACKTRACE ===============
+ backtrace #0: hit 1, time   2.109 ms
+   [0] main (0x400570)
 
+========== FUNCTION CALL GRAPH ==========
+   2.109 ms : (1) main
+   2.109 ms : (1) foo
+   2.098 ms : (1) bar
+   2.096 ms : (1) usleep
 """, sort='graph')
 
     def pre(self):
