@@ -11,12 +11,12 @@ class TestCase(TestBase):
    0.714 us [ 4435] | __monstartup();
    0.349 us [ 4435] | __cxa_atexit();
             [ 4435] | main() {
-            [ 4435] |   open() {
+            [ 4435] |   fopen() {
    6.413 us [ 4435] |     sys_open();
-   7.037 us [ 4435] |   } /* open */
-            [ 4435] |   close() {
+   7.037 us [ 4435] |   } /* fopen */
+            [ 4435] |   fclose() {
    8.389 us [ 4435] |     sys_close();
-   9.949 us [ 4435] |   } /* close */
+   9.949 us [ 4435] |   } /* fclose */
   17.632 us [ 4435] | } /* main */
 """)
 
@@ -33,3 +33,6 @@ class TestCase(TestBase):
         # while -N option works on record time and accept a glob
         return '%s -K3 -T %s@kernel,depth=1 -N %s@kernel -N %s@kernel %s' % \
             (TestBase.ftrace, '^sys_', 'exit_to_usermode_loop', 'smp_irq_work_interrupt', 't-' + self.name)
+
+    def fixup(self, cflags, result):
+        return result.replace('sys_open', 'sys_openat')
