@@ -1,5 +1,4 @@
 #include <stdio.h>
-#include <string.h>
 #include <stdint.h>
 #include <inttypes.h>
 #include <unistd.h>
@@ -7,16 +6,21 @@
 #include <errno.h>
 #include <limits.h>
 #include <assert.h>
+#include <byteswap.h>
 
 #include "uftrace.h"
 #include "utils/utils.h"
 #include "utils/fstack.h"
+#include "utils/filter.h"
 #include "utils/symbol.h"
+#include "utils/kernel.h"
+#include "utils/perf.h"
 #include "libmcount/mcount.h"
 
 
 /**
  * read_task_file - read 'task' file from data directory
+ * @sess: session link to manage sessions and tasks
  * @dirname: name of the data directory
  * @needs_session: read session info too
  * @sym_rel_addr: whether symbol address is relative
@@ -28,14 +32,16 @@
  *
  * It returns 0 for success, -1 for error.
  */
-int read_task_file(char *dirname, bool needs_session, bool sym_rel_addr)
+int read_task_file(struct uftrace_session_link *sess, char *dirname,
+		   bool needs_session, bool sym_rel_addr)
 {
 	int fd;
 	char pad[8];
 	char buf[1024];
-	struct ftrace_msg msg;
-	struct ftrace_msg_task task;
-	struct ftrace_msg_sess sess;
+	struct uftrace_msg msg;
+	struct uftrace_msg_task tmsg;
+	struct uftrace_msg_sess smsg;
+	int ret = -1;
 
 	snprintf(buf, sizeof(buf), "%s/task", dirname);
 	fd = open(buf, O_RDONLY);
@@ -44,52 +50,52 @@ int read_task_file(char *dirname, bool needs_session, bool sym_rel_addr)
 
 	pr_dbg("reading task file\n");
 	while (read_all(fd, &msg, sizeof(msg)) == 0) {
-		if (msg.magic != FTRACE_MSG_MAGIC)
-			return -1;
+		if (msg.magic != UFTRACE_MSG_MAGIC)
+			goto out;
 
 		switch (msg.type) {
-		case FTRACE_MSG_SESSION:
-			if (read_all(fd, &sess, sizeof(sess)) < 0)
-				return -1;
-			if (read_all(fd, buf, sess.namelen) < 0)
-				return -1;
-			if (sess.namelen % 8 &&
-			    read_all(fd, pad, 8 - (sess.namelen % 8)) < 0)
-				return -1;
+		case UFTRACE_MSG_SESSION:
+			if (read_all(fd, &smsg, sizeof(smsg)) < 0)
+				goto out;
+			if (read_all(fd, buf, smsg.namelen) < 0)
+				goto out;
+			if (smsg.namelen % 8 &&
+			    read_all(fd, pad, 8 - (smsg.namelen % 8)) < 0)
+				goto out;
 
 			if (needs_session)
-				create_session(&sess, dirname, buf, sym_rel_addr);
+				create_session(sess, &smsg, dirname, buf, sym_rel_addr);
 			break;
 
-		case FTRACE_MSG_TID:
-			if (read_all(fd, &task, sizeof(task)) < 0)
-				return -1;
+		case UFTRACE_MSG_TASK_START:
+			if (read_all(fd, &tmsg, sizeof(tmsg)) < 0)
+				goto out;
 
-			create_task(&task, false, needs_session);
+			create_task(sess, &tmsg, false, needs_session);
 			break;
 
-		case FTRACE_MSG_FORK_END:
-			if (read_all(fd, &task, sizeof(task)) < 0)
-				return -1;
+		case UFTRACE_MSG_FORK_END:
+			if (read_all(fd, &tmsg, sizeof(tmsg)) < 0)
+				goto out;
 
-			create_task(&task, true, needs_session);
+			create_task(sess, &tmsg, true, needs_session);
 			break;
 
 		default:
-			pr_log("invalid contents in task file\n");
-			return -1;
+			pr_dbg("invalid contents in task file\n");
+			goto out;
 		}
 	}
+	ret = 0;
 
-	if (needs_session)
-		set_kernel_base(dirname, first_session->sid);
-
+out:
 	close(fd);
-	return 0;
+	return ret;
 }
 
 /**
  * read_task_txt_file - read 'task.txt' file from data directory
+ * @sess: session link to manage sessions and tasks
  * @dirname: name of the data directory
  * @needs_session: read session info too
  * @sym_rel_addr: whethere symbol address is relative
@@ -99,16 +105,17 @@ int read_task_file(char *dirname, bool needs_session, bool sym_rel_addr)
  *
  * It returns 0 for success, -1 for error.
  */
-int read_task_txt_file(char *dirname, bool needs_session, bool sym_rel_addr)
+int read_task_txt_file(struct uftrace_session_link *sess, char *dirname,
+		       bool needs_session, bool sym_rel_addr)
 {
 	FILE *fp;
 	char *fname = NULL;
 	char *line = NULL;
 	size_t sz = 0;
 	long sec, nsec;
-	struct ftrace_msg_task task;
-	struct ftrace_msg_sess sess;
-	struct ftrace_msg_dlopen dlop;
+	struct uftrace_msg_task tmsg;
+	struct uftrace_msg_sess smsg;
+	struct uftrace_msg_dlopen dlop;
 	char *exename, *pos;
 
 	xasprintf(&fname, "%s/%s", dirname, "task.txt");
@@ -123,24 +130,24 @@ int read_task_txt_file(char *dirname, bool needs_session, bool sym_rel_addr)
 	while (getline(&line, &sz, fp) >= 0) {
 		if (!strncmp(line, "TASK", 4)) {
 			sscanf(line + 5, "timestamp=%lu.%lu tid=%d pid=%d",
-			       &sec, &nsec, &task.tid, &task.pid);
+			       &sec, &nsec, &tmsg.tid, &tmsg.pid);
 
-			task.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
-			create_task(&task, false, needs_session);
+			tmsg.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
+			create_task(sess, &tmsg, false, needs_session);
 		}
 		else if (!strncmp(line, "FORK", 4)) {
 			sscanf(line + 5, "timestamp=%lu.%lu pid=%d ppid=%d",
-			       &sec, &nsec, &task.tid, &task.pid);
+			       &sec, &nsec, &tmsg.tid, &tmsg.pid);
 
-			task.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
-			create_task(&task, true, needs_session);
+			tmsg.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
+			create_task(sess, &tmsg, true, needs_session);
 		}
 		else if (!strncmp(line, "SESS", 4)) {
 			if (!needs_session)
 				continue;
 
 			sscanf(line + 5, "timestamp=%lu.%lu %*[^i]id=%d sid=%s",
-			       &sec, &nsec, &sess.task.pid, (char *)&sess.sid);
+			       &sec, &nsec, &smsg.task.pid, (char *)&smsg.sid);
 
 			// Get the execname
 			pos = strstr(line, "exename=");
@@ -151,14 +158,14 @@ int read_task_txt_file(char *dirname, bool needs_session, bool sym_rel_addr)
 			if (pos)
 				*pos = '\0';
 
-			sess.task.tid = sess.task.pid;
-			sess.task.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
-			sess.namelen = strlen(exename);
+			smsg.task.tid = smsg.task.pid;
+			smsg.task.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
+			smsg.namelen = strlen(exename);
 
-			create_session(&sess, dirname, exename, sym_rel_addr);
+			create_session(sess, &smsg, dirname, exename, sym_rel_addr);
 		}
 		else if (!strncmp(line, "DLOP", 4)) {
-			struct ftrace_session *s;
+			struct uftrace_session *s;
 
 			if (!needs_session)
 				continue;
@@ -179,15 +186,65 @@ int read_task_txt_file(char *dirname, bool needs_session, bool sym_rel_addr)
 			dlop.task.time = (uint64_t)sec * NSEC_PER_SEC + nsec;
 			dlop.namelen = strlen(exename);
 
-			s = get_session_from_sid(dlop.sid);
+			s = get_session_from_sid(sess, dlop.sid);
 			assert(s);
-			session_add_dlopen(s, dirname, dlop.task.time,
+			session_add_dlopen(s, dlop.task.time,
 					   dlop.base_addr, exename);
 		}
 	}
 
-	if (needs_session)
-		set_kernel_base(dirname, first_session->sid);
+	fclose(fp);
+	free(fname);
+	return 0;
+}
+
+/**
+ * read_events_file - read 'events.txt' file from data directory
+ * @dirname: name of the data directory
+ *
+ * This function read the events file in the @dirname and build event
+ * information (for userspace).
+ *
+ * It returns 0 for success, -1 for error.
+ */
+int read_events_file(struct ftrace_file_handle *handle)
+{
+	FILE *fp;
+	char *fname = NULL;
+	char *line = NULL;
+	size_t sz = 0;
+
+	xasprintf(&fname, "%s/%s", handle->dirname, "events.txt");
+
+	fp = fopen(fname, "r");
+	if (fp == NULL) {
+		/* it might hit no events, so no file is ok */
+		if (errno == ENOENT)
+			errno = 0;
+
+		free(fname);
+		return -errno;
+	}
+
+	pr_dbg("reading %s file\n", fname);
+	while (getline(&line, &sz, fp) >= 0) {
+		char provider[512];
+		char event[512];
+		unsigned evt_id;
+		struct uftrace_event *ev;
+
+		if (!strncmp(line, "EVENT", 5)) {
+			sscanf(line + 7, "%u %[^:]:%s",
+			       &evt_id, provider, event);
+
+			ev = xmalloc(sizeof(*ev));
+			ev->id = evt_id;
+			ev->provider = xstrdup(provider);
+			ev->event = xstrdup(event);
+
+			list_add_tail(&ev->list, &handle->events);
+		}
+	}
 
 	fclose(fp);
 	free(fname);
@@ -200,7 +257,7 @@ static void snprint_timestamp(char *buf, size_t sz, uint64_t timestamp)
 		 timestamp / NSEC_PER_SEC, timestamp % NSEC_PER_SEC);
 }
 
-void write_task_info(const char *dirname, struct ftrace_msg_task *tmsg)
+void write_task_info(const char *dirname, struct uftrace_msg_task *tmsg)
 {
 	FILE *fp;
 	char *fname = NULL;
@@ -220,7 +277,7 @@ void write_task_info(const char *dirname, struct ftrace_msg_task *tmsg)
 	free(fname);
 }
 
-void write_fork_info(const char *dirname, struct ftrace_msg_task *tmsg)
+void write_fork_info(const char *dirname, struct uftrace_msg_task *tmsg)
 {
 	FILE *fp;
 	char *fname = NULL;
@@ -240,7 +297,7 @@ void write_fork_info(const char *dirname, struct ftrace_msg_task *tmsg)
 	free(fname);
 }
 
-void write_session_info(const char *dirname, struct ftrace_msg_sess *smsg,
+void write_session_info(const char *dirname, struct uftrace_msg_sess *smsg,
 			const char *exename)
 {
 	FILE *fp;
@@ -261,7 +318,7 @@ void write_session_info(const char *dirname, struct ftrace_msg_sess *smsg,
 	free(fname);
 }
 
-void write_dlopen_info(const char *dirname, struct ftrace_msg_dlopen *dmsg,
+void write_dlopen_info(const char *dirname, struct uftrace_msg_dlopen *dmsg,
 		       const char *libname)
 {
 	FILE *fp;
@@ -285,7 +342,7 @@ void write_dlopen_info(const char *dirname, struct ftrace_msg_dlopen *dmsg,
 static void check_data_order(struct ftrace_file_handle *handle)
 {
 	union {
-		struct ftrace_ret_stack s;
+		struct uftrace_record s;
 		uint64_t d[2];
 	} data;
 
@@ -301,74 +358,80 @@ static void check_data_order(struct ftrace_file_handle *handle)
 		pr_dbg("bitfield order is different!\n");
 }
 
-#define RECORD_MSG  "Was '%s' compiled with -pg or\n"		\
-"\t-finstrument-functions flag and ran with ftrace record?\n"
-
 int open_data_file(struct opts *opts, struct ftrace_file_handle *handle)
 {
 	int ret = -1;
 	FILE *fp;
 	char buf[PATH_MAX];
-	bool again = false;
 
 	snprintf(buf, sizeof(buf), "%s/info", opts->dirname);
 
-retry:
 	fp = fopen(buf, "rb");
-	if (fp == NULL) {
-		if (again) {
-			/* restore original file name for error reporting */
-			snprintf(buf, sizeof(buf), "%s/info", opts->dirname);
+	if (fp != NULL)
+		goto ok;
+
+	/* if default dirname is failed */
+	if (!strcmp(opts->dirname, UFTRACE_DIR_NAME)) {
+		/* try again inside the current directory */
+		fp = fopen("./info", "rb");
+		if (fp != NULL) {
+			opts->dirname = "./";
+			goto ok;
 		}
 
-		if (errno == ENOENT) {
-			if (!again && !strcmp(opts->dirname, UFTRACE_DIR_NAME)) {
-				/* retry with old default dirname */
-				snprintf(buf, sizeof(buf), "%s/info",
-					UFTRACE_DIR_OLD_NAME);
-
-				again = true;
-				goto retry;
-			}
-
-			pr_log("cannot find %s file!\n", buf);
-
-			if (opts->exename)
-				pr_err(RECORD_MSG, opts->exename);
-		} else {
-			pr_err("cannot open %s file", buf);
+		/* retry with old default dirname */
+		snprintf(buf, sizeof(buf), "%s/info", UFTRACE_DIR_OLD_NAME);
+		fp = fopen(buf, "rb");
+		if (fp != NULL) {
+			opts->dirname = UFTRACE_DIR_OLD_NAME;
+			goto ok;
 		}
-		goto out;
+
+		/* restore original file name for error reporting */
+		snprintf(buf, sizeof(buf), "%s/info", opts->dirname);
 	}
 
-	if (again) {
-		/* found data in old dirname, rename it */
-		opts->dirname = UFTRACE_DIR_OLD_NAME;
-	}
+	/* data file loading is failed */
+	pr_dbg("cannot open %s file\n", buf);
+	goto out;
 
+ok:
 	handle->fp = fp;
 	handle->dirname = opts->dirname;
 	handle->depth = opts->depth;
-	handle->kern = NULL;
 	handle->nr_tasks = 0;
 	handle->tasks = NULL;
 	handle->time_filter = opts->threshold;
 	handle->time_range = opts->range;
+	handle->sessions.root  = RB_ROOT;
+	handle->sessions.tasks = RB_ROOT;
+	handle->sessions.first = NULL;
+	handle->kernel = NULL;
+	handle->nr_perf = 0;
+	handle->perf = NULL;
+	INIT_LIST_HEAD(&handle->events);
 
 	if (fread(&handle->hdr, sizeof(handle->hdr), 1, fp) != 1)
 		pr_err("cannot read header data");
 
 	if (memcmp(handle->hdr.magic, UFTRACE_MAGIC_STR, UFTRACE_MAGIC_LEN))
-		pr_err("invalid magic string found!");
-
-	if (handle->hdr.version < UFTRACE_FILE_VERSION_MIN ||
-	    handle->hdr.version > UFTRACE_FILE_VERSION)
-		pr_err("unsupported file version: %u", handle->hdr.version);
+		pr_err_ns("invalid magic string found!");
 
 	check_data_order(handle);
 
-	if (read_ftrace_info(handle->hdr.info_mask, handle) < 0)
-		pr_err("cannot read ftrace header info!");
+	if (handle->needs_byte_swap) {
+		handle->hdr.version   = bswap_32(handle->hdr.version);
+		handle->hdr.feat_mask = bswap_64(handle->hdr.feat_mask);
+		handle->hdr.info_mask = bswap_64(handle->hdr.info_mask);
+		handle->hdr.max_stack = bswap_16(handle->hdr.max_stack);
+	}
+
+	if (handle->hdr.version < UFTRACE_FILE_VERSION_MIN ||
+	    handle->hdr.version > UFTRACE_FILE_VERSION)
+		pr_err_ns("unsupported file version: %u", handle->hdr.version);
+
+	if (read_uftrace_info(handle->hdr.info_mask, handle) < 0)
+		pr_err_ns("cannot read uftrace header info!");
 
 	fclose(fp);
 
@@ -377,22 +440,65 @@ retry:
 
 	if (handle->hdr.feat_mask & TASK_SESSION) {
 		bool sym_rel = false;
+		struct uftrace_session_link *sessions = &handle->sessions;
 
 		if (handle->hdr.feat_mask & SYM_REL_ADDR)
 			sym_rel = true;
 
-		// read task.txt first and then try old task file
-		if (read_task_txt_file(opts->dirname, true, sym_rel) < 0 &&
-		    read_task_file(opts->dirname, true, sym_rel) < 0)
-			pr_warn("invalid task file\n");
+		/* read old task file first and then try task.txt file */
+		if (read_task_file(sessions, opts->dirname, true, sym_rel) < 0 &&
+		    read_task_txt_file(sessions, opts->dirname, true, sym_rel) < 0) {
+			if (errno == ENOENT)
+				pr_warn("no trace data found\n");
+			else
+				pr_warn("invalid trace data\n");
+		}
 	}
 
-	if (handle->hdr.feat_mask & (ARGUMENT | RETVAL))
-		setup_fstack_args(handle->info.argspec);
+	if (handle->hdr.info_mask & ARG_SPEC) {
+		if (handle->hdr.feat_mask & AUTO_ARGS) {
+			setup_auto_args_str(handle->info.autoarg,
+					    handle->info.autoret,
+					    handle->info.autoenum);
+		}
+
+		setup_fstack_args(handle->info.argspec, handle->info.retspec,
+				  handle, false);
+		if (handle->info.auto_args_enabled) {
+			setup_fstack_args(handle->info.autoarg,
+					  handle->info.autoret,
+					  handle, true);
+		}
+	}
 
 	if (!(handle->hdr.feat_mask & MAX_STACK))
 		handle->hdr.max_stack = MCOUNT_RSTACK_MAX;
 
+	if (handle->hdr.feat_mask & KERNEL) {
+		struct uftrace_kernel_reader *kernel;
+
+		kernel = xzalloc(sizeof(*kernel));
+
+		kernel->handle   = handle;
+		kernel->dirname  = opts->dirname;
+		kernel->skip_out = opts->kernel_skip_out;
+
+		if (setup_kernel_data(kernel) == 0) {
+			handle->kernel = kernel;
+			load_kernel_symbol(opts->dirname);
+		}
+		else {
+			free(kernel);
+			handle->kernel = NULL;
+		}
+	}
+
+	if (handle->hdr.feat_mask & EVENT)
+		read_events_file(handle);
+
+	if (handle->hdr.feat_mask & PERF_EVENT)
+		setup_perf_data(handle);
+
 	ret = 0;
 
 out:
@@ -404,6 +510,19 @@ void close_data_file(struct opts *opts, struct ftrace_file_handle *handle)
 	if (opts->exename == handle->info.exename)
 		opts->exename = NULL;
 
-	clear_ftrace_info(&handle->info);
+	if (has_kernel_data(handle->kernel)) {
+		finish_kernel_data(handle->kernel);
+		free(handle->kernel);
+	}
+
+	if (has_perf_data(handle))
+		finish_perf_data(handle);
+
+	delete_sessions(&handle->sessions);
+
+	if (handle->hdr.feat_mask & AUTO_ARGS)
+		finish_auto_args();
+
+	clear_uftrace_info(&handle->info);
 	reset_task_handle(handle);
 }
