@@ -1,7 +1,7 @@
 /*
  * Very simple (and incomplete by design) C++ name demangler.
  *
- * Copyright (C) 2015-2016, LG Electronics, Namhyung Kim <namhyung.kim@lge.com>
+ * Copyright (C) 2015-2017, LG Electronics, Namhyung Kim <namhyung.kim@lge.com>
  *
  * Released under the GPL v2.
  *
@@ -9,8 +9,6 @@
  */
 
 #include <stdlib.h>
-#include <string.h>
-#include <ctype.h>
 #include <assert.h>
 
 /* This should be defined before #include "utils.h" */
@@ -1077,7 +1075,7 @@ static int dd_special_name(struct demangle_data *dd)
 {
 	char c0 = dd_curr(dd);
 	char c1 = dd_peek(dd, 1);
-	char T_type[] = "VTIS";
+	char T_type[] = "VTISFJ";
 
 	if (dd_eof(dd))
 		return -1;
@@ -1104,10 +1102,12 @@ static int dd_special_name(struct demangle_data *dd)
 	}
 	if (c0 == 'G') {
 		if (c1 == 'V') {
+			/* guard */
 			dd_consume_n(dd, 2);
 			return dd_name(dd);
 		}
 		if (c1 == 'R') {
+			/* reftemp */
 			dd_consume_n(dd, 2);
 			if (dd_name(dd) < 0)
 				return -1;
@@ -1117,6 +1117,23 @@ static int dd_special_name(struct demangle_data *dd)
 			__DD_DEBUG_CONSUME(dd, '_');
 			return 0;
 		}
+		if (c1 == 'A') {
+			/* hidden alias */
+			dd_consume_n(dd, 2);
+			return dd_encoding(dd);
+		}
+		if (c1 == 'T') {
+			dd_consume_n(dd, 2);
+
+			c0 = dd_curr(dd);
+			/* (non-)transaction clone */
+			if (c0 == 't' || c0 == 'n') {
+				dd_consume(dd);
+				return dd_encoding(dd);
+			}
+
+			return -1;
+		}
 	}
 
 	DD_DEBUG(dd, "valid special name", 0);
@@ -1344,7 +1361,7 @@ static int dd_local_name(struct demangle_data *dd)
 	c = dd_curr(dd);
 	if (c == 'd') {
 		__dd_consume(dd, NULL);
-		if (dd_number(dd) < 0)
+		if (dd_curr(dd) != '_' && dd_number(dd) < 0)
 			return -1;
 		__DD_DEBUG_CONSUME(dd, '_');
 		if (dd_name(dd) < 0)
@@ -1392,6 +1409,7 @@ static int dd_encoding(struct demangle_data *dd)
 {
 	int ret;
 	char c = dd_curr(dd);
+	char end[] = "E.@";
 
 	if (dd_eof(dd))
 		return -1;
@@ -1405,7 +1423,7 @@ static int dd_encoding(struct demangle_data *dd)
 	if (ret < 0)
 		return ret;
 
-	while (!dd_eof(dd) && dd_curr(dd) != 'E' && dd_curr(dd) != '.') {
+	while (!dd_eof(dd) && !strchr(end, dd_curr(dd))) {
 		__dd_add_debug(dd, "dd_type");
 
 		if (dd_type(dd) < 0)
@@ -1415,6 +1433,9 @@ static int dd_encoding(struct demangle_data *dd)
 	/* ignore compiler generated suffix: XXX.part.0 */
 	if (dd_curr(dd) == '.')
 		dd->len = dd->pos;
+	/* ignore version info in PLT symbols: malloc@GLIBC2.1 */
+	if (dd_curr(dd) == '@')
+		dd->len = dd->pos;
 
 	return 0;
 }
@@ -1473,16 +1494,6 @@ static char *demangle_full(char *str)
  */
 char *demangle(char *str)
 {
-	static const size_t size_of_gsi = sizeof("_GLOBAL__sub_I") - 1;
-
-	/* skip global initialize (constructor?) functions */
-	if (strncmp(str, "_GLOBAL__sub_I", size_of_gsi) == 0) {
-		str += size_of_gsi;
-
-		while (*str++ != '_')
-			continue;
-	}
-
 	switch (demangler) {
 	case DEMANGLE_SIMPLE:
 		return demangle_simple(str);
@@ -1499,113 +1510,175 @@ char *demangle(char *str)
 #ifdef UNIT_TEST
 TEST_CASE(demangle_simple1)
 {
+	char *name;
+
 	dbg_domain[DBG_DEMANGLE] = 2;
 
-	TEST_STREQ("normal", demangle_simple("normal"));
-	TEST_STREQ("ABC::foo", demangle_simple("_ZN3ABC3fooEv"));
-	TEST_STREQ("ABC::ABC", demangle_simple("_ZN3ABCC1Ei"));
-	TEST_STREQ("operator new", demangle_simple("_Znwm"));
-	TEST_STREQ("ns::ns1::foo::bar1", demangle_simple("_ZN2ns3ns13foo4bar1Ev"));
+	name = demangle_simple("normal");
+	TEST_STREQ("normal", name);
+	free(name);
+
+	name = demangle_simple("_ZN3ABC3fooEv");
+	TEST_STREQ("ABC::foo", name);
+	free(name);
+
+	name = demangle_simple("_ZN3ABCC1Ei");
+	TEST_STREQ("ABC::ABC", name);
+	free(name);
+
+	name = demangle_simple("_Znwm");
+	TEST_STREQ("operator new", name);
+	free(name);
+
+	name = demangle_simple("_ZN2ns3ns13foo4bar1Ev");
+	TEST_STREQ("ns::ns1::foo::bar1", name);
+	free(name);
 
 	return TEST_OK;
 }
 
 TEST_CASE(demangle_simple2)
 {
+	char *name;
+
 	dbg_domain[DBG_DEMANGLE] = 2;
 
-	TEST_STREQ("FtraceService::~FtraceService",
-		   demangle_simple("_ZThn8_N13FtraceServiceD0Ev"));
-	TEST_STREQ("v8::internal::ScopedVector::ScopedVector",
-		   demangle_simple("_ZN2v88internal12ScopedVectorIcEC1Ei"));
-	TEST_STREQ("std::allocator_traits::construct",
-		   demangle_simple("_ZNSt16allocator_traitsISaISt13_Rb_tree_node"
-				   "ISt4pairIKSsN7pbnjson7JSchemaEEEEE9construct"
-				   "IS6_IS1_ISsS4_EEEEDTcl12_S_constructfp_fp0_"
-				   "spcl7forwardIT0_Efp1_EEERS7_PT_DpOSB_"));
+	name = demangle_simple("_ZThn8_N13FtraceServiceD0Ev");
+	TEST_STREQ("FtraceService::~FtraceService", name);
+	free(name);
+
+	name = demangle_simple("_ZN2v88internal12ScopedVectorIcEC1Ei");
+	TEST_STREQ("v8::internal::ScopedVector::ScopedVector", name);
+	free(name);
+
+	name = demangle_simple("_ZNSt16allocator_traitsISaISt13_Rb_tree_node"
+			       "ISt4pairIKSsN7pbnjson7JSchemaEEEEE9construct"
+			       "IS6_IS1_ISsS4_EEEEDTcl12_S_constructfp_fp0_"
+			       "spcl7forwardIT0_Efp1_EEERS7_PT_DpOSB_");
+	TEST_STREQ("std::allocator_traits::construct", name);
+	free(name);
 
 	return TEST_OK;
 }
 
 TEST_CASE(demangle_simple3)
 {
+	char *name;
+
 	dbg_domain[DBG_DEMANGLE] = 2;
 
-	TEST_STREQ("node::Watchdog::Destroy",
-		   demangle_simple("_ZN4node8Watchdog7DestroyEv.part.0"));
-	TEST_STREQ("v8::internal::CodeStub::GetKey",
-		   demangle_simple("_ZN2v88internal8CodeStub6GetKeyEv.constprop.17"));
-	TEST_STREQ("std::operator ==",
-		   demangle_simple("_ZSteqIPN2v88internal8compiler4NodeERKS4_PS5_E"
-				   "bRKSt15_Deque_iteratorIT_T0_T1_ESE_"));
-	TEST_STREQ("v8::base::internal::operator *",
-		   demangle_simple("_ZN2v84base8internalmlIiiEENS1_14CheckedNumeric"
-				   "INS1_19ArithmeticPromotionIT_T0_XqugtsrNS1_"
-				   "11MaxExponentIS5_EE5valuesrNS7_IS6_EE5value"
-				   "qugtsrS8_5valueL_ZNS7_IiE5valueEELNS1_"
-				   "27ArithmeticPromotionCategoryE0ELSB_2E"
-				   "qugtsrS9_5valueL_ZNSA_5valueEELSB_1ELSB_2EEE"
-				   "4typeEEERKNS3_IS5_EES6_"));
-	TEST_STREQ("std::pow",
-		   demangle_simple("_ZSt3powIidEN9__gnu_cxx11__promote_2IT_T0_NS0_"
-				   "9__promoteIS2_XsrSt12__is_integerIS2_E7__valueEE"
-				   "6__typeENS4_IS3_XsrS5_IS3_E7__valueEE6__typeEE"
-				   "6__typeES2_S3_"));
+	name = demangle_simple("_ZN4node8Watchdog7DestroyEv.part.0");
+	TEST_STREQ("node::Watchdog::Destroy", name);
+	free(name);
+
+	name = demangle_simple("_ZN2v88internal8CodeStub6GetKeyEv.constprop.17");
+	TEST_STREQ("v8::internal::CodeStub::GetKey", name);
+	free(name);
+
+	name = demangle_simple("_ZSteqIPN2v88internal8compiler4NodeERKS4_PS5_E"
+			       "bRKSt15_Deque_iteratorIT_T0_T1_ESE_");
+	TEST_STREQ("std::operator ==", name);
+	free(name);
+
+	name = demangle_simple("_ZN2v84base8internalmlIiiEENS1_14CheckedNumeric"
+			       "INS1_19ArithmeticPromotionIT_T0_XqugtsrNS1_"
+			       "11MaxExponentIS5_EE5valuesrNS7_IS6_EE5value"
+			       "qugtsrS8_5valueL_ZNS7_IiE5valueEELNS1_"
+			       "27ArithmeticPromotionCategoryE0ELSB_2E"
+			       "qugtsrS9_5valueL_ZNSA_5valueEELSB_1ELSB_2EEE"
+			       "4typeEEERKNS3_IS5_EES6_");
+	TEST_STREQ("v8::base::internal::operator *", name);
+	free(name);
+
+	name = demangle_simple("_ZSt3powIidEN9__gnu_cxx11__promote_2IT_T0_NS0_"
+			       "9__promoteIS2_XsrSt12__is_integerIS2_E7__valueEE"
+			       "6__typeENS4_IS3_XsrS5_IS3_E7__valueEE6__typeEE"
+			       "6__typeES2_S3_");
+	TEST_STREQ("std::pow", name);
+	free(name);
 
 	return TEST_OK;
 }
 
 TEST_CASE(demangle_simple4)
 {
+	char *name;
+
 	dbg_domain[DBG_DEMANGLE] = 2;
 
-	TEST_STREQ("std::__find_if",
-		   demangle_simple("_ZSt9__find_ifISt14_List_iteratorISt10shared_ptr"
-				   "I16AppLaunchingItemEEZN13MemoryChecker8add_itemE"
-				   "S1_I13LaunchingItemEEUlS7_E_ET_S9_S9_T0_"
-				   "St18input_iterator_tag"));
+	name = demangle_simple("_ZSt9__find_ifISt14_List_iteratorISt10shared_ptr"
+			       "I16AppLaunchingItemEEZN13MemoryChecker8add_itemE"
+			       "S1_I13LaunchingItemEEUlS7_E_ET_S9_S9_T0_"
+			       "St18input_iterator_tag");
+	TEST_STREQ("std::__find_if", name);
+	free(name);
 
-	TEST_STREQ("convertToWindowType::~convertToWindowType",
-		   demangle_simple("_ZZ19convertToWindowTypeRKSsRSsENUt_D1Ev"));
+	name = demangle_simple("_ZZ19convertToWindowTypeRKSsRSsENUt_D1Ev");
+	TEST_STREQ("convertToWindowType::~convertToWindowType", name);
+	free(name);
 
-	TEST_STREQ("std::set::erase::cxx11",
-		   demangle_simple("_ZNSt3setISsSt4lessISsESaISsEE5eraseB5cxx11E"
-				   "St23_Rb_tree_const_iteratorISsE"));
+	name = demangle_simple("_ZNSt3setISsSt4lessISsESaISsEE5eraseB5cxx11E"
+			       "St23_Rb_tree_const_iteratorISsE");
+	TEST_STREQ("std::set::erase::cxx11", name);
+	free(name);
 
-	TEST_STREQ("std::allocator_traits::_S_select",
-		   demangle_simple("_ZNSt16allocator_traitsISaISsEE9_S_select"
-				   "IKS0_EENSt9enable_ifIXntsrNS1_15__select_helper"
-				   "IT_EE5valueES6_E4typeERS6_"));
+	name = demangle_simple("_ZNSt16allocator_traitsISaISsEE9_S_select"
+			       "IKS0_EENSt9enable_ifIXntsrNS1_15__select_helper"
+			       "IT_EE5valueES6_E4typeERS6_");
+	TEST_STREQ("std::allocator_traits::_S_select", name);
+	free(name);
 
-	TEST_STREQ("icu_54::umtx_loadAcquire",
-		   demangle_simple("_ZN6icu_5416umtx_loadAcquireERU7_Atomici"));
+	name = demangle_simple("_ZN6icu_5416umtx_loadAcquireERU7_Atomici");
+	TEST_STREQ("icu_54::umtx_loadAcquire", name);
+	free(name);
 
 	return TEST_OK;
 }
 
 TEST_CASE(demangle_simple5)
 {
+	char *name;
+
 	dbg_domain[DBG_DEMANGLE] = 2;
 
-	TEST_STREQ("v8::internal::RememberedSet::Iterate",
-		   demangle_simple("_ZN2v88internal13RememberedSetILNS0_"
-				   "16PointerDirectionE1EE7IterateIZNS3_"
-				   "18IterateWithWrapperIPFvPPNS0_10HeapObjectE"
-				   "S7_EEEvPNS0_4HeapET_EUlPhE_EEvSC_SD_"));
+	name = demangle_simple("_ZN2v88internal13RememberedSetILNS0_"
+			       "16PointerDirectionE1EE7IterateIZNS3_"
+			       "18IterateWithWrapperIPFvPPNS0_10HeapObjectE"
+			       "S7_EEEvPNS0_4HeapET_EUlPhE_EEvSC_SD_");
+	TEST_STREQ("v8::internal::RememberedSet::Iterate", name);
+	free(name);
 
-	TEST_STREQ("v8::internal::SlotSet::Iterate",
-		   demangle_simple("_ZN2v88internal7SlotSet7Iterate"
-				   "IZNS0_13RememberedSetILNS0_16PointerDirectionE"
-				   "1EE18IterateWithWrapperIPFvPPNS0_10HeapObjectE"
-				   "S8_EEEvPNS0_4HeapET_EUlPhE_EEiSE_"));
+	name = demangle_simple("_ZN2v88internal7SlotSet7Iterate"
+			       "IZNS0_13RememberedSetILNS0_16PointerDirectionE"
+			       "1EE18IterateWithWrapperIPFvPPNS0_10HeapObjectE"
+			       "S8_EEEvPNS0_4HeapET_EUlPhE_EEiSE_");
+	TEST_STREQ("v8::internal::SlotSet::Iterate", name);
+	free(name);
 
-	TEST_STREQ("std::tuple::tuple",
-		   demangle_simple("_ZNSt5tupleIJPbSt14default_deleteIA_bEEEC2Ev"));
+	name = demangle_simple("_ZNSt5tupleIJPbSt14default_deleteIA_bEEEC2Ev");
+	TEST_STREQ("std::tuple::tuple", name);
+	free(name);
 
-	TEST_STREQ("storageIndexFromLayoutItem",
-		   demangle_simple("_Z26storageIndexFromLayoutItemRK"
-				   "N51_GLOBAL__N_kernel_qformlayout.cpp_C3DE8A26_2E30FA86"
-				   "17FixedColumnMatrixIP15QFormLayoutItemLi2EEES2_"));
+	name = demangle_simple("_Z26storageIndexFromLayoutItemRK"
+			       "N51_GLOBAL__N_kernel_qformlayout.cpp_C3DE8A26_2E30FA86"
+			       "17FixedColumnMatrixIP15QFormLayoutItemLi2EEES2_");
+	TEST_STREQ("storageIndexFromLayoutItem", name);
+	free(name);
+
+	name = demangle_simple("_ZGTtNSt11range_errorD1Ev");
+	TEST_STREQ("std::range_error::~range_error", name);
+	free(name);
+
+	name = demangle_simple("_ZNSi6ignoreEl@@GLIBCXX_3.4.5");
+	TEST_STREQ("std::basic_istream::ignore", name);
+	free(name);
+
+	name = demangle_simple("_ZN4llvm12function_refIFN5clang12ActionResult"
+			       "IPNS1_4ExprELb1EEES4_EE11callback_fnIZNS1_4Sema"
+			       "25CorrectDelayedTyposInExprES4_PNS1_7VarDeclE"
+			       "S7_Ed_NUlS4_E_EEES5_lS4_");
+	TEST_STREQ("llvm::function_ref::callback_fn", name);
+	free(name);
 
 	return TEST_OK;
 }
