@@ -1,5 +1,5 @@
-#ifndef __FTRACE_FSTACK_H__
-#define __FTRACE_FSTACK_H__
+#ifndef UFTRACE_FSTACK_H
+#define UFTRACE_FSTACK_H
 
 #include <stdio.h>
 #include <stdint.h>
@@ -8,7 +8,7 @@
 #include "../uftrace.h"
 
 struct sym;
-struct ftrace_trigger;
+struct uftrace_trigger;
 
 enum fstack_flag {
 	FSTACK_FL_FILTERED	= (1U << 0),
@@ -36,22 +36,25 @@ struct ftrace_task_handle {
 	bool valid;
 	bool done;
 	bool lost_seen;
+	bool fork_handled;
 	bool fstack_set;
 	bool display_depth_set;
 	FILE *fp;
 	struct sym *func;
-	struct ftrace_task *t;
+	struct uftrace_task *t;
 	struct ftrace_file_handle *h;
-	struct ftrace_ret_stack ustack;
-	struct ftrace_ret_stack kstack;
-	struct ftrace_ret_stack *rstack;
+	struct uftrace_record ustack;
+	struct uftrace_record kstack;
+	struct uftrace_record *rstack;
 	struct uftrace_rstack_list rstack_list;
 	int stack_count;
 	int lost_count;
 	int user_stack_count;
 	int display_depth;
 	int user_display_depth;
+	int fork_display_depth;
 	int column_index;
+	int event_color;
 	enum context ctx;
 	uint64_t timestamp;
 	uint64_t timestamp_last;
@@ -109,27 +112,38 @@ void fstack_consume(struct ftrace_file_handle *handle,
 int read_task_ustack(struct ftrace_file_handle *handle,
 		     struct ftrace_task_handle *task);
 int read_task_args(struct ftrace_task_handle *task,
-		   struct ftrace_ret_stack *rstack,
+		   struct uftrace_record *rstack,
 		   bool is_retval);
 
+static inline bool is_user_record(struct ftrace_task_handle *task,
+				  struct uftrace_record *rec)
+{
+	return rec == &task->ustack;
+}
+
+static inline bool is_kernel_record(struct ftrace_task_handle *task,
+				    struct uftrace_record *rec)
+{
+	return rec == &task->kstack;
+}
+
 void setup_task_filter(char *tid_filter, struct ftrace_file_handle *handle);
-int setup_fstack_filters(char *filter_str, char *trigger_str);
-void setup_fstack_args(char *argspec);
-void fstack_prepare_fixup(void);
+void setup_fstack_args(char *argspec, char *retspec,
+		       struct ftrace_file_handle *handle, bool auto_args);
 int fstack_setup_filters(struct opts *opts, struct ftrace_file_handle *handle);
 
 int fstack_entry(struct ftrace_task_handle *task,
-		 struct ftrace_ret_stack *rstack,
-		 struct ftrace_trigger *tr);
+		 struct uftrace_record *rstack,
+		 struct uftrace_trigger *tr);
 void fstack_exit(struct ftrace_task_handle *task);
 int fstack_update(int type, struct ftrace_task_handle *task,
 		  struct fstack *fstack);
 struct ftrace_task_handle *fstack_skip(struct ftrace_file_handle *handle,
 				       struct ftrace_task_handle *task,
-				       int curr_depth);
+				       int curr_depth, bool event_skip_out);
 bool fstack_check_filter(struct ftrace_task_handle *task);
 void get_argspec_string(struct ftrace_task_handle *task,
 		        char *args, size_t len,
 		        enum argspec_string_bits str_mode);
 
-#endif /* __FTRACE_FSTACK_H__ */
+#endif /* UFTRACE_FSTACK_H */
