@@ -1,20 +1,23 @@
 /*
- * utiltily functions and macros for ftrace
+ * utiltily functions and macros for uftrace
  *
- * Copyright (C) 2014-2016, LG Electronics, Namhyung Kim <namhyung.kim@lge.com>
+ * Copyright (C) 2014-2017, LG Electronics, Namhyung Kim <namhyung.kim@lge.com>
  *
  * Released under the GPL v2.
  */
 
-#ifndef __FTRACE_UTILS_H__
-#define __FTRACE_UTILS_H__
+#ifndef UFTRACE_UTILS_H
+#define UFTRACE_UTILS_H
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
 #include <stdbool.h>
 #include <endian.h>
+#include <string.h>
+#include <ctype.h>
 
+#include "compiler.h"
 
 #ifndef container_of
 # define container_of(ptr, type, member) ({			\
@@ -43,8 +46,9 @@ extern int debug;
 extern FILE *logfp;
 extern FILE *outfp;
 
+/* must change DBG_DOMAIN_STR (in mcount.h) as well */
 enum debug_domain {
-	DBG_FTRACE	= 0,
+	DBG_UFTRACE	= 0,
 	DBG_SYMBOL,
 	DBG_DEMANGLE,
 	DBG_FILTER,
@@ -52,6 +56,9 @@ enum debug_domain {
 	DBG_SESSION,
 	DBG_KERNEL,
 	DBG_MCOUNT,
+	DBG_DYNAMIC,
+	DBG_EVENT,
+	DBG_SCRIPT,
 	DBG_DOMAIN_MAX,
 };
 extern int dbg_domain[DBG_DOMAIN_MAX];
@@ -72,8 +79,9 @@ enum color_setting {
 #define COLOR_CODE_GRAY     'g'
 #define COLOR_CODE_BOLD     'b'
 
+#define DEFAULT_EVENT_COLOR  COLOR_CODE_GRAY
+
 extern void __pr_dbg(const char *fmt, ...);
-extern void __pr_log(const char *fmt, ...);
 extern void __pr_out(const char *fmt, ...);
 extern void __pr_err(const char *fmt, ...) __attribute__((noreturn));
 extern void __pr_err_s(const char *fmt, ...) __attribute__((noreturn));
@@ -90,7 +98,7 @@ extern void setup_signal(void);
 #endif
 
 #ifndef PR_DOMAIN
-# define PR_DOMAIN  DBG_FTRACE
+# define PR_DOMAIN  DBG_UFTRACE
 #endif
 
 #define pr_dbg(fmt, ...) 					\
@@ -111,10 +119,6 @@ extern void setup_signal(void);
 		__pr_dbg(PR_FMT ": " fmt, ## __VA_ARGS__);	\
 })
 
-#define pr_log(fmt, ...)					\
-	__pr_log(PR_FMT ": %s:%d:%s\n" fmt,			\
-		 __FILE__, __LINE__, __func__, ## __VA_ARGS__)
-
 #define pr_err(fmt, ...)					\
 	__pr_err_s(PR_FMT ": %s:%d:%s\n ERROR: " fmt,		\
 		 __FILE__, __LINE__, __func__, ## __VA_ARGS__)
@@ -125,9 +129,9 @@ extern void setup_signal(void);
 
 #define pr_warn(fmt, ...)	__pr_warn("WARN: " fmt, ## __VA_ARGS__)
 
-#define pr_cont(fmt, ...)	__pr_log(fmt, ## __VA_ARGS__)
 #define pr_out(fmt, ...)	__pr_out(fmt, ## __VA_ARGS__)
-#define pr_use(fmt, ...)	__pr_out(fmt, ## __VA_ARGS__)
+#define pr_cont(fmt, ...)	__pr_out(fmt, ## __VA_ARGS__)
+#define pr_use(fmt, ...)	__pr_out("Usage: " fmt, ## __VA_ARGS__)
 
 #define pr_red(fmt, ...)	__pr_color(COLOR_CODE_RED,     fmt, ## __VA_ARGS__)
 #define pr_green(fmt, ...)	__pr_color(COLOR_CODE_GREEN,   fmt, ## __VA_ARGS__)
@@ -224,6 +228,7 @@ struct uftrace_time_range {
 struct iovec;
 
 int read_all(int fd, void *buf, size_t size);
+int pread_all(int fd, void *buf, size_t size, off_t off);
 int fread_all(void *byf, size_t size, FILE *fp);
 int write_all(int fd, void *buf, size_t size);
 int writev_all(int fd, struct iovec *iov, int count);
@@ -235,10 +240,25 @@ char *read_exename(void);
 
 void print_time_unit(uint64_t delta_nsec);
 void print_diff_percent(uint64_t base_nsec, uint64_t delta_nsec);
+void print_diff_time_unit(uint64_t base_nsec, uint64_t pair_nsec);
+void print_diff_count(unsigned long base, unsigned long pair);
 
 void start_pager(void);
 void wait_for_pager(void);
 
 bool check_time_range(struct uftrace_time_range *range, uint64_t timestamp);
+uint64_t parse_time(char *arg, int limited_digits);
 
-#endif /* __FTRACE_UTILS_H__ */
+char * strjoin(char *left, char *right, char *delim);
+char * strquote(char *str, int *len);
+
+char **parse_cmdline(char *cmd, int *argc);
+void free_parsed_cmdline(char **argv);
+
+struct ftrace_file_handle;
+
+char *get_event_name(struct ftrace_file_handle *handle, unsigned evt_id);
+
+char *absolute_dirname(const char *path, char *resolved_path);
+
+#endif /* UFTRACE_UTILS_H */
